from typing import Dict, Set, Type

from lightning.pytorch.cli import LightningArgumentParser, LightningCLI

from chebai.preprocessing.datasets.base import XYBaseDataModule
from chebai.trainer.CustomTrainer import CustomTrainer


class ChebaiCLI(LightningCLI):
    """
    Custom CLI subclass for Chebai project based on PyTorch Lightning's LightningCLI.

    Args:
        save_config_kwargs (dict): Keyword arguments for saving configuration.
        parser_kwargs (dict): Keyword arguments for parser configuration.

    Attributes:
        save_config_kwargs (dict): Configuration options for saving.
        parser_kwargs (dict): Configuration options for the argument parser.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize ChebaiCLI with custom trainer and configure parser settings.

        Args:
            args (list): List of arguments for LightningCLI.
            kwargs (dict): Keyword arguments for LightningCLI.
                save_config_kwargs (dict): Keyword arguments for saving configuration.
                parser_kwargs (dict): Keyword arguments for parser configuration.
        """
        super().__init__(trainer_class=CustomTrainer, *args, **kwargs)

    def add_arguments_to_parser(self, parser: LightningArgumentParser):
        """
        Link input parameters that are used by different classes (e.g. number of labels)
        see https://lightning.ai/docs/pytorch/stable/cli/lightning_cli_expert.html#argument-linking

        Args:
            parser (LightningArgumentParser): Argument parser instance.
        """

        def call_data_methods(data: Type[XYBaseDataModule]):
            if data._num_of_labels is None:
                data.prepare_data()
                data.setup()
            return data.num_of_labels

        parser.link_arguments(
            "data",
            "model.init_args.out_dim",
            apply_on="instantiate",
            compute_fn=call_data_methods,
        )

        parser.link_arguments(
            "data.feature_vector_size",
            "model.init_args.input_dim",
            apply_on="instantiate",
        )

        for kind in ("train", "val", "test"):
            for average in (
                "micro-f1",
                "macro-f1",
                "balanced-accuracy",
                "roc-auc",
                "f1",
                "mse",
                "rmse",
                "r2",
            ):
                # When using lightning > 2.5.1 then need to uncomment all metrics that are not used
                # for average in ("mse", "rmse","r2"): # for regression
                # for average in ("f1", "roc-auc"): # for binary classification
                # for average in ("micro-f1", "macro-f1", "roc-auc"): # for multilabel classification
                # for average in ("micro-f1", "macro-f1", "balanced-accuracy", "roc-auc"): # for multilabel classification using balanced-accuracy
                parser.link_arguments(
                    "data.num_of_labels",
                    f"model.init_args.{kind}_metrics.init_args.metrics.{average}.init_args.num_labels",
                    apply_on="instantiate",
                )

        parser.link_arguments(
            "data.num_of_labels", "trainer.callbacks.init_args.num_labels"
        )
        # parser.link_arguments(
        #     "model.init_args.out_dim", "trainer.callbacks.init_args.num_labels"
        # )
        # parser.link_arguments(
        #     "data", "model.init_args.criterion.init_args.data_extractor"
        # )
        # parser.link_arguments(
        #     "data.init_args.chebi_version",
        #     "model.init_args.criterion.init_args.data_extractor.init_args.chebi_version",
        # )

        parser.link_arguments(
            "data", "model.init_args.criterion.init_args.data_extractor"
        )

    @staticmethod
    def subcommands() -> Dict[str, Set[str]]:
        """
        Defines the list of available subcommands and the arguments to skip.

        Returns:
            Dict[str, Set[str]]: Dictionary where keys are subcommands and values are sets of arguments to skip.
        """
        return {
            "fit": {"model", "train_dataloaders", "val_dataloaders", "datamodule"},
            "validate": {"model", "dataloaders", "datamodule"},
            "test": {"model", "dataloaders", "datamodule"},
            "predict": {"model", "dataloaders", "datamodule"},
            "predict_from_file": {"model"},
        }


def cli():
    """
    Main function to instantiate and run the ChebaiCLI.
    """
    ChebaiCLI(
        save_config_kwargs={"config_filename": "lightning_config.yaml"},
        parser_kwargs={"parser_mode": "omegaconf"},
    )



import json

from chebai.result.base import ResultProcessor


class JSONResultProcessor(ResultProcessor):
    @classmethod
    def _identifier(cls):
        return "json"

    def start(self):
        self.data = []

    def close(self):
        with open("predictions.json", "w") as fout:
            json.dump(self.data, fout)
            del self.data

    def process_prediction(self, proc_id, raw_features, labels, preds, ident, **kwargs):
        self.data.append(
            dict(
                ident=ident,
                labels=labels if labels is not None else None,
                prediction=preds.tolist(),
            )
        )



import gc
import os
import traceback
from datetime import datetime
from pathlib import Path
from typing import List, LiteralString, Optional, Tuple

import pandas as pd
import torch
import wandb
from torchmetrics.functional.classification import (
    multilabel_auroc,
    multilabel_average_precision,
    multilabel_f1_score,
)

from chebai.loss.semantic import DisjointLoss
from chebai.models import Electra
from chebai.preprocessing.datasets.base import _DynamicDataset
from chebai.preprocessing.datasets.chebi import ChEBIOver100
from chebai.result.utils import (
    evaluate_model,
    get_checkpoint_from_wandb,
    load_results_from_buffer,
)

DEVICE = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


def binary(left, right):
    return torch.logical_and(left > 0.5, right > 0.5)


def strict(left, right):
    return left + right > 1


def weak(left, right):
    return left + right > 1.01


def product(left, right):
    return left * right


def lukasiewicz(left, right):
    return torch.relu(left + right - 1)


def apply_metric(metric, left, right):
    return torch.sum(metric(left, right), dim=0)


ALL_CONSISTENCY_METRICS = [product, lukasiewicz, weak, strict, binary]


def _filter_to_one_hot(preds, idx_filter):
    """Takes list of indices (e.g. [1, 3, 0]) and returns a one-hot filter with these indices
    (e.g. [[0,1,0,0], [0,0,0,1], [1,0,0,0]])"""
    res = torch.zeros((len(idx_filter), preds.shape[1]), dtype=torch.bool)
    for i, idx in enumerate(idx_filter):
        res[i][idx] = True
    return res


def _sort_results_by_label(n_labels, results, filter):
    by_label = torch.zeros(n_labels, device=DEVICE, dtype=torch.int)
    for r, filter_l in zip(results, filter):
        by_label[filter_l] += r
    return by_label


def get_best_epoch(run):
    files = run.files()
    best_ep = None
    best_micro_f1 = 0
    for file in files:
        if file.name.startswith("checkpoints/best_epoch"):
            micro_f1 = float(file.name.split("=")[-1][:-5])
            if micro_f1 > best_micro_f1 or best_ep is None:
                best_ep = int(file.name.split("=")[1].split("_")[0])
                best_micro_f1 = micro_f1
    if best_ep is None:
        raise Exception(f"Could not find any 'best' checkpoint for run {run.id}")
    else:
        print(f"Best epoch for run {run.id}: {best_ep}")
    return best_ep


def download_model_from_wandb(
    run_id, base_dir=os.path.join("logs", "downloaded_ckpts")
):
    api = wandb.Api()
    run = api.run(f"chebai/chebai/{run_id}")
    epoch = get_best_epoch(run)
    return (
        get_checkpoint_from_wandb(epoch, run, root=base_dir),
        epoch,
    )


def load_preds_labels(
    ckpt_path: LiteralString, data_module, data_subset_key="test", buffer_dir=None
):
    if buffer_dir is None:
        buffer_dir = os.path.join(
            "results_buffer",
            *ckpt_path.split(os.path.sep)[-2:],
            f"{data_module.__class__.__name__}_{data_subset_key}",
        )
    model = Electra.load_from_checkpoint(ckpt_path, map_location="cuda:0", strict=False)
    print(
        f"Calculating predictions on {data_module.__class__.__name__} ({data_subset_key})..."
    )
    evaluate_model(
        model,
        data_module,
        buffer_dir=buffer_dir,
        # for chebi, use kinds, otherwise use file names
        filename=(
            data_subset_key if not isinstance(buffer_dir, _DynamicDataset) else None
        ),
        kind=data_subset_key,
        skip_existing_preds=True,
        batch_size=1,
    )
    return load_results_from_buffer(buffer_dir, device=torch.device("cpu"))


def get_label_names(data_module):
    if os.path.exists(os.path.join(data_module.processed_dir_main, "classes.txt")):
        with open(os.path.join(data_module.processed_dir_main, "classes.txt")) as fin:
            return [line.strip() for line in fin]
    print(
        f"Failed to retrieve label names, {os.path.join(data_module.processed_dir_main, 'classes.txt')} not found"
    )
    return None


def get_chebi_graph(data_module, label_names):
    if os.path.exists(os.path.join(data_module.raw_dir, "chebi.obo")):
        chebi_graph = data_module._extract_class_hierarchy(
            os.path.join(data_module.raw_dir, "chebi.obo")
        )
        if label_names is None:
            return chebi_graph
        return chebi_graph.subgraph([int(n) for n in label_names])
    print(
        f"Failed to retrieve ChEBI graph, {os.path.join(data_module.raw_dir, 'chebi.obo')} not found"
    )
    return None


def get_disjoint_groups(disjoint_files):
    if disjoint_files is None:
        disjoint_files = os.path.join("data", "chebi-disjoints.owl")
    disjoint_pairs, disjoint_groups = [], []
    for file in disjoint_files:
        if isinstance(file, Path):
            file = str(file)
        if file.endswith(".csv"):
            disjoint_pairs += pd.read_csv(file, header=None).values.tolist()
        elif file.endswith(".owl"):
            with open(file, "r") as f:
                plaintext = f.read()
                segments = plaintext.split("<")
                disjoint_pairs = []
                left = None
                for seg in segments:
                    if seg.startswith("rdf:Description ") or seg.startswith(
                        "owl:Class"
                    ):
                        left = int(seg.split('rdf:about="&obo;CHEBI_')[1].split('"')[0])
                    elif seg.startswith("owl:disjointWith"):
                        right = int(
                            seg.split('rdf:resource="&obo;CHEBI_')[1].split('"')[0]
                        )
                        disjoint_pairs.append([left, right])

                disjoint_groups = []
                for seg in plaintext.split("<rdf:Description>"):
                    if "owl;AllDisjointClasses" in seg:
                        classes = seg.split('rdf:about="&obo;CHEBI_')[1:]
                        classes = [int(c.split('"')[0]) for c in classes]
                        disjoint_groups.append(classes)
        else:
            raise NotImplementedError(
                "Unsupported disjoint file format: " + file.split(".")[-1]
            )

    disjoint_all = disjoint_pairs + disjoint_groups
    # one disjointness is commented out in the owl-file
    # (the correct way would be to parse the owl file and notice the comment symbols, but for this case, it should work)
    if [22729, 51880] in disjoint_all:
        disjoint_all.remove([22729, 51880])
    # print(f"Found {len(disjoint_all)} disjoint groups")
    return disjoint_all


class PredictionSmoother:
    """Removes implication and disjointness violations from predictions"""

    def __init__(self, dataset, label_names=None, disjoint_files=None):
        self.chebi_graph = get_chebi_graph(dataset, None)
        self.set_label_names(label_names)
        self.disjoint_groups = get_disjoint_groups(disjoint_files)

    def set_label_names(self, label_names):
        if label_names is not None:
            self.label_names = [int(label) for label in label_names]
            chebi_subgraph = self.chebi_graph.subgraph(self.label_names)
            self.label_successors = torch.zeros(
                (len(self.label_names), len(self.label_names)), dtype=torch.bool
            )
            for i, label in enumerate(self.label_names):
                self.label_successors[i, i] = 1
                for p in chebi_subgraph.successors(label):
                    if p in self.label_names:
                        self.label_successors[i, self.label_names.index(p)] = 1
            self.label_successors = self.label_successors.unsqueeze(0)

    def __call__(self, preds):
        if preds.shape[1] == 0:
            # no labels predicted
            return preds
        # preds shape: (n_samples, n_labels)
        preds_sum_orig = torch.sum(preds)
        # step 1: apply implications: for each class, set prediction to max of itself and all successors
        preds = preds.unsqueeze(1)
        preds_masked_succ = torch.where(self.label_successors, preds, 0)
        # preds_masked_succ shape: (n_samples, n_labels, n_labels)

        preds = preds_masked_succ.max(dim=2).values
        if torch.sum(preds) != preds_sum_orig:
            print(f"Preds change (step 1): {torch.sum(preds) - preds_sum_orig}")
        preds_sum_orig = torch.sum(preds)
        # step 2: eliminate disjointness violations: for group of disjoint classes, set all except max to 0.49 (if it is not already lower)
        preds_bounded = torch.min(preds, torch.ones_like(preds) * 0.49)
        for disj_group in self.disjoint_groups:
            disj_group = [
                self.label_names.index(g) for g in disj_group if g in self.label_names
            ]
            if len(disj_group) > 1:
                old_preds = preds[:, disj_group]
                disj_max = torch.max(preds[:, disj_group], dim=1)
                for i, row in enumerate(preds):
                    for l_ in range(len(preds[i])):
                        if l_ in disj_group and l_ != disj_group[disj_max.indices[i]]:
                            preds[i, l_] = preds_bounded[i, l_]
                samples_changed = 0
                for i, row in enumerate(preds[:, disj_group]):
                    if any(r != o for r, o in zip(row, old_preds[i])):
                        samples_changed += 1
                if samples_changed != 0:
                    print(
                        f"disjointness group {[self.label_names[d] for d in disj_group]} changed {samples_changed} samples"
                    )
        if torch.sum(preds) != preds_sum_orig:
            print(f"Preds change (step 2): {torch.sum(preds) - preds_sum_orig}")
        preds_sum_orig = torch.sum(preds)
        # step 3: disjointness violation removal may have caused new implication inconsistencies -> set each prediction to min of predecessors
        preds = preds.unsqueeze(1)
        preds_masked_predec = torch.where(
            torch.transpose(self.label_successors, 1, 2), preds, 1
        )
        preds = preds_masked_predec.min(dim=2).values
        if torch.sum(preds) != preds_sum_orig:
            print(f"Preds change (step 3): {torch.sum(preds) - preds_sum_orig}")
        return preds


def _filter_to_dense(filter):
    filter_dense = []
    for i in range(filter.shape[0]):
        for j in range(filter.shape[1]):
            if filter[i, j] > 0:
                filter_dense.append([i, j])
    return torch.tensor(filter_dense)


def build_prediction_filter(data_module_labeled=None):
    if data_module_labeled is None:
        data_module_labeled = ChEBIOver100(chebi_version=231)
    # prepare filters
    print("Loading implication / disjointness filters...")
    dl = DisjointLoss(
        path_to_disjointness=os.path.join("data", "disjoint.csv"),
        data_extractor=data_module_labeled,
    )
    impl = _filter_to_dense(dl.implication_filter_l)
    disj = _filter_to_dense(dl.disjoint_filter_l)

    return [
        (impl[:, 0], impl[:, 1], "impl"),
        (disj[:, 0], disj[:, 1], "disj"),
    ]


def run_consistency_metrics(
    preds,
    consistency_filters,
    data_module_labeled=None,  # use labels from this dataset for violations
    violation_metrics=None,
    verbose_violation_output=False,
    save_details_to=None,
):
    """Calculates all semantic metrics for given predictions (and supervised metrics if labels are provided)"""
    if violation_metrics is None:
        violation_metrics = ALL_CONSISTENCY_METRICS
    if data_module_labeled is None:
        data_module_labeled = ChEBIOver100(chebi_version=231)
    if save_details_to is not None:
        os.makedirs(save_details_to, exist_ok=True)

    preds.to("cpu")

    n_labels = preds.size(1)
    print(f"Found {preds.shape[0]} predictions ({n_labels} classes)")

    results = {}

    for dl_filter_l, dl_filter_r, filter_type in consistency_filters:
        l_preds = preds[:, dl_filter_l]
        r_preds = preds[:, dl_filter_r]
        for i, metric in enumerate(violation_metrics):
            if metric.__name__ not in results:
                results[metric.__name__] = {}
            print(f"Calculating metrics {metric.__name__} on {filter_type}")

            metric_results = {}
            metric_results["tps"] = torch.sum(
                torch.stack(
                    [
                        apply_metric(
                            metric,
                            l_preds[i : i + 1000],
                            (
                                r_preds[i : i + 1000]
                                if filter_type == "impl"
                                else 1 - r_preds[i : i + 1000]
                            ),
                        )
                        for i in range(0, r_preds.shape[0], 1000)
                    ]
                ),
                dim=0,
            )
            metric_results["fns"] = torch.sum(
                torch.stack(
                    [
                        apply_metric(
                            metric,
                            l_preds[i : i + 1000],
                            (
                                1 - r_preds[i : i + 1000]
                                if filter_type == "impl"
                                else r_preds[i : i + 1000]
                            ),
                        )
                        for i in range(0, r_preds.shape[0], 1000)
                    ]
                ),
                dim=0,
            )
            if verbose_violation_output:
                label_names = get_label_names(data_module_labeled)
                print(
                    f"Found {torch.sum(metric_results['fns'])} {filter_type}-violations"
                )
                # for k, fn_cls in enumerate(metric_results['fns']):
                #    if fn_cls > 0:
                #        print(f"\tThereof, {fn_cls.item()} belong to class {label_names[k]}")
                if torch.sum(metric_results["fns"]) != 0:
                    fns = metric(
                        l_preds, 1 - r_preds if filter_type == "impl" else r_preds
                    )
                    print(fns.shape)
                    for k, row in enumerate(fns):
                        if torch.sum(row) != 0:
                            print(f"{torch.sum(row)} violations for entity {k}")
                            for j, violation in enumerate(row):
                                if violation > 0:
                                    print(
                                        f"\tviolated ({label_names[dl_filter_l[j]]} -> {preds[k, dl_filter_l[j]]:.3f}"
                                        f", {label_names[dl_filter_r[j]]} -> {preds[k, dl_filter_r[j]]:.3f})"
                                    )

            m_l_agg = {}
            for key, value in metric_results.items():
                m_l_agg[key] = _sort_results_by_label(
                    n_labels,
                    value,
                    dl_filter_l,
                )
            m_r_agg = {}
            for key, value in metric_results.items():
                m_r_agg[key] = _sort_results_by_label(
                    n_labels,
                    value,
                    dl_filter_r,
                )

            if save_details_to is not None:
                with open(
                    os.path.join(
                        save_details_to, f"{metric.__name__}_{filter_type}_all.csv"
                    ),
                    "w+",
                ) as f:
                    f.write("left,right,tps,fns\n")
                    for left, right, tps, fns in zip(
                        dl_filter_l,
                        dl_filter_r,
                        metric_results["tps"],
                        metric_results["fns"],
                    ):
                        f.write(f"{left},{right},{tps},{fns}\n")
                with open(
                    os.path.join(
                        save_details_to, f"{metric.__name__}_{filter_type}_l.csv"
                    ),
                    "w+",
                ) as f:
                    f.write("left,tps,fns\n")
                    for left in range(n_labels):
                        f.write(
                            f"{left},{m_l_agg['tps'][left].item()},{m_l_agg['fns'][left].item()}\n"
                        )
                with open(
                    os.path.join(
                        save_details_to, f"{metric.__name__}_{filter_type}_r.csv"
                    ),
                    "w+",
                ) as f:
                    f.write("right,tps,fns\n")
                    for right in range(n_labels):
                        f.write(
                            f"{right},{m_r_agg['tps'][right].item()},{m_r_agg['fns'][right].item()}\n"
                        )
                print(
                    f"Saved unaggregated consistency metrics ({metric.__name__}, {filter_type}) to {save_details_to}"
                )

            fns_sum = torch.sum(metric_results["fns"]).item()
            results[metric.__name__][f"micro-fnr-{filter_type}"] = (
                0
                if fns_sum == 0
                else (
                    torch.sum(metric_results["fns"])
                    / (
                        torch.sum(metric_results["tps"])
                        + torch.sum(metric_results["fns"])
                    )
                ).item()
            )
            macro_fnr_l = m_l_agg["fns"] / (m_l_agg["tps"] + m_l_agg["fns"])
            results[metric.__name__][f"lmacro-fnr-{filter_type}"] = (
                0
                if fns_sum == 0
                else torch.mean(macro_fnr_l[~macro_fnr_l.isnan()]).item()
            )
            macro_fnr_r = m_r_agg["fns"] / (m_r_agg["tps"] + m_r_agg["fns"])
            results[metric.__name__][f"rmacro-fnr-{filter_type}"] = (
                0
                if fns_sum == 0
                else torch.mean(macro_fnr_r[~macro_fnr_r.isnan()]).item()
            )
            results[metric.__name__][f"fn-sum-{filter_type}"] = torch.sum(
                metric_results["fns"]
            ).item()
            results[metric.__name__][f"tp-sum-{filter_type}"] = torch.sum(
                metric_results["tps"]
            ).item()

            del metric_results
            del m_l_agg
            del m_r_agg

            gc.collect()
        del l_preds
        del r_preds
        gc.collect()

    return results


def run_supervised_metrics(preds, labels, save_details_to=None):
    # calculate supervised metrics
    results = {}
    if labels is not None:
        results["micro-f1"] = multilabel_f1_score(
            preds, labels, num_labels=preds.size(1), average="micro"
        ).item()
        results["macro-f1"] = multilabel_f1_score(
            preds, labels, num_labels=preds.size(1), average="macro"
        ).item()
        results["micro-roc-auc"] = multilabel_auroc(
            preds, labels, num_labels=preds.size(1), average="micro"
        ).item()
        results["macro-roc-auc"] = multilabel_auroc(
            preds, labels, num_labels=preds.size(1), average="macro"
        ).item()

        results["micro-ap"] = multilabel_average_precision(
            preds, labels, num_labels=preds.size(1), average="micro"
        ).item()
        results["macro-ap"] = multilabel_average_precision(
            preds, labels, num_labels=preds.size(1), average="macro"
        ).item()

        if save_details_to is not None:
            f1_by_label = multilabel_f1_score(
                preds, labels, num_labels=preds.size(1), average=None
            )
            roc_by_label = multilabel_auroc(
                preds, labels, num_labels=preds.size(1), average=None
            )
            ap_by_label = multilabel_average_precision(
                preds, labels, num_labels=preds.size(1), average=None
            )
            with open(os.path.join(save_details_to, "supervised.csv"), "w+") as f:
                f.write("label,f1,roc-auc,ap\n")
                for right in range(preds.size(1)):
                    f.write(
                        f"{right},{f1_by_label[right].item()},{roc_by_label[right].item()},{ap_by_label[right].item()}\n"
                    )
            print(f"Saved class-wise supervised metrics to {save_details_to}")

    del preds
    del labels
    gc.collect()
    return results


# run predictions / metrics calculations for semantic loss paper runs (NeSy 2024 submission)
def run_semloss_eval():
    # runs from wandb
    non_wandb_runs = []
    api = wandb.Api()
    runs = api.runs("chebai/chebai", filters={"tags": "eval_semloss_paper"})
    print(f"Found {len(runs)} tagged wandb runs")
    # ids_wandb = [run.id for run in runs]

    # ids used in the NeSy submission
    prod = ["tk15yznc", "uke62a8m", "w0h3zr5s"]
    xu19 = ["5ko8knb4", "061fd85t", "r50ioujs"]
    prod_mixed = ["hk8555ff", "e0lxw8py", "lig23cmg"]
    luka = ["0c0s48nh", "lfg384bp", "qeghvubh"]
    baseline = ["i4wtz1k4", "zd020wkv", "rc1q3t49"]
    prodk2 = ["ng3usn0p", "rp0wwzjv", "8fma1q7r"]
    ids = baseline + prod + prodk2 + xu19 + luka + prod_mixed
    # ids = ids_wandb
    run_all(
        ids,
        non_wandb_runs,
        prediction_datasets=[(ChEBIOver100(chebi_version=231), "test")],
        consistency_metrics=[binary],
    )


def run_all(
    wandb_ids=None,
    local_ckpts: List[Tuple] = None,
    consistency_metrics: Optional[List[callable]] = None,
    prediction_datasets: List[Tuple] = None,
    remove_violations: bool = False,
    results_dir="_fuzzy_loss_eval",
    check_consistency_on=None,
    verbose_violation_output=False,
):
    if wandb_ids is None:
        wandb_ids = []
    if local_ckpts is None:
        local_ckpts = []
    if consistency_metrics is None:
        consistency_metrics = ALL_CONSISTENCY_METRICS
    if prediction_datasets is None:
        prediction_datasets = [
            (ChEBIOver100(chebi_version=231), "test"),
        ]
    if check_consistency_on is None:
        check_consistency_on = ChEBIOver100(chebi_version=231)

    if remove_violations:
        smooth_preds = PredictionSmoother(check_consistency_on)
    else:
        smooth_preds = lambda x: x  # noqa: E731

    timestamp = datetime.now().strftime("%y%m%d-%H%M%S")
    prediction_filters = build_prediction_filter(check_consistency_on)

    results_path_consistency = os.path.join(
        results_dir,
        f"consistency_metrics_{timestamp}{'_violations_removed' if remove_violations else ''}.csv",
    )
    consistency_keys = [
        "micro-fnr-impl",
        "lmacro-fnr-impl",
        "rmacro-fnr-impl",
        "fn-sum-impl",
        "tp-sum-impl",
        "micro-fnr-disj",
        "lmacro-fnr-disj",
        "rmacro-fnr-disj",
        "fn-sum-disj",
        "tp-sum-disj",
    ]
    with open(results_path_consistency, "x") as f:
        f.write(
            "run-id,epoch,datamodule,data_key,metric,"
            + ",".join(consistency_keys)
            + "\n"
        )
    results_path_supervised = os.path.join(
        results_dir,
        f"supervised_metrics_{timestamp}{'_violations_removed' if remove_violations else ''}.csv",
    )
    supervised_keys = [
        "micro-f1",
        "macro-f1",
        "micro-roc-auc",
        "macro-roc-auc",
        "micro-ap",
        "macro-ap",
    ]
    with open(results_path_supervised, "x") as f:
        f.write("run-id,epoch,datamodule,data_key," + ",".join(supervised_keys) + "\n")

    ckpts = [(run_name, ep, None) for run_name, ep in local_ckpts] + [
        (None, None, wandb_id) for wandb_id in wandb_ids
    ]

    for run_name, epoch, wandb_id in ckpts:
        try:
            ckpt_dir = os.path.join("logs", "downloaded_ckpts")
            # for wandb runs, use short id as name, otherwise use ckpt dir name
            if wandb_id is not None:
                run_name = wandb_id
                ckpt_path, epoch = download_model_from_wandb(run_name, ckpt_dir)
            else:
                ckpt_path = None
                for file in os.listdir(os.path.join(ckpt_dir, run_name)):
                    if f"epoch={epoch}_" in file or f"epoch={epoch}." in file:
                        ckpt_path = os.path.join(os.path.join(ckpt_dir, run_name, file))
                assert (
                    ckpt_path is not None
                ), f"Failed to find checkpoint for epoch {epoch} in {os.path.join(ckpt_dir, run_name)}"
            print(f"Starting run {run_name} (epoch {epoch})")

            for dataset, dataset_key in prediction_datasets:
                # copy data from legacy buffer dir if possible
                old_buffer_dir = os.path.join(
                    "results_buffer",
                    *ckpt_path.split(os.path.sep)[-2:],
                    f"{dataset.__class__.__name__}_{dataset_key}",
                )
                buffer_dir = os.path.join(
                    "results_buffer",
                    run_name,
                    f"epoch={epoch}",
                    f"{dataset.__class__.__name__}_{dataset_key}",
                )
                print("Checking for buffer dir", old_buffer_dir)
                if os.path.isdir(old_buffer_dir):
                    from distutils.dir_util import copy_tree, remove_tree

                    os.makedirs(buffer_dir, exist_ok=True)
                    copy_tree(old_buffer_dir, buffer_dir)
                    remove_tree(old_buffer_dir, dry_run=True)
                    print(f"Moved buffer from {old_buffer_dir} to {buffer_dir}")
                print(f"Using buffer_dir {buffer_dir}")
                preds, labels = load_preds_labels(
                    ckpt_path, dataset, dataset_key, buffer_dir
                )
                # identity function if remove_violations is False
                smooth_preds(preds)

                details_path = None  # os.path.join(
                #    results_dir,
                #    f"{run_name}_ep{epoch}_{dataset.__class__.__name__}_{dataset_key}",
                # )
                metrics_dict = run_consistency_metrics(
                    preds,
                    prediction_filters,
                    check_consistency_on,
                    consistency_metrics,
                    verbose_violation_output,
                    save_details_to=details_path,
                )
                with open(results_path_consistency, "a") as f:
                    for metric in metrics_dict:
                        values = metrics_dict[metric]
                        f.write(
                            f"{run_name},{epoch},{dataset.__class__.__name__},{dataset_key},{metric},"
                            f"{','.join([str(values[k]) for k in consistency_keys])}\n"
                        )
                print(
                    f"Consistency metrics have been written to {results_path_consistency}"
                )
                if labels is not None:
                    metrics_dict = run_supervised_metrics(
                        preds, labels, save_details_to=details_path
                    )
                    with open(results_path_supervised, "a") as f:
                        f.write(
                            f"{run_name},{epoch},{dataset.__class__.__name__},{dataset_key},"
                            f"{','.join([str(metrics_dict[k]) for k in supervised_keys])}\n"
                        )
                    print(
                        f"Supervised metrics have been written to {results_path_supervised}"
                    )
        except Exception as e:
            print(
                f"Error during run {wandb_id if wandb_id is not None else run_name}: {e}"
            )
            print(traceback.format_exc())


# follow-up to NeSy submission
def run_fuzzy_loss(tag="fuzzy_loss", skip_first_n=0):
    api = wandb.Api()
    runs = api.runs("chebai/chebai", filters={"tags": tag})
    print(f"Found {len(runs)} wandb runs tagged with '{tag}'")
    ids = [run.id for run in runs]
    chebi100 = ChEBIOver100(
        chebi_version=231,
        splits_file_path=os.path.join(
            "data", "chebi_v231", "ChEBI100", "fuzzy_loss_splits.csv"
        ),
    )
    local_ckpts = [][skip_first_n:]
    # pubchem_kmeans = PubChemKMeans()
    run_all(
        ids[max(0, skip_first_n - len(local_ckpts)) :],  # ids,
        local_ckpts,
        consistency_metrics=[binary],
        check_consistency_on=chebi100,
        prediction_datasets=[
            (chebi100, "test"),
            # (pubchem_kmeans, "cluster1_cutoff2k.pt"),
            # (pubchem_kmeans, "cluster2.pt"),
            # (pubchem_kmeans, "ten_from_each_cluster.pt"),
            # (pubchem_kmeans, "chebi_close.pt"),
        ],
    )


if __name__ == "__main__":
    import sys

    if len(sys.argv) > 2:
        run_fuzzy_loss(sys.argv[1], int(sys.argv[2]))
    elif len(sys.argv) > 1:
        run_fuzzy_loss(sys.argv[1])
    else:
        run_fuzzy_loss()



__all__ = [
    "JCIData",
    "JCIExtendedTokenData",
    "JCIExtendedBPEData",
    "JCIExtSelfies",
    "JCITokenData",
    "ChEBIOver100",
    "JCI_500_COLUMNS",
    "JCI_500_COLUMNS_INT",
]

import os
import pickle
import random
from abc import ABC
from collections import OrderedDict
from itertools import cycle, permutations, product
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Literal, Optional, Union

import numpy as np
import pandas as pd
import torch
from rdkit import Chem

from chebai.preprocessing import reader as dr
from chebai.preprocessing.datasets.base import XYBaseDataModule, _DynamicDataset

if TYPE_CHECKING:
    import fastobo
    import networkx as nx

# exclude some entities from the dataset because the violate disjointness axioms
CHEBI_BLACKLIST = [
    194026,
    144321,
    156504,
    167175,
    167174,
    167178,
    183506,
    74635,
    3311,
    190439,
    92386,
]


class JCIBase(XYBaseDataModule):
    LABEL_INDEX = 2
    SMILES_INDEX = 1

    @property
    def _name(self):
        return "JCI"

    @property
    def processed_file_names(self):
        return ["test.pt", "train.pt", "validation.pt"]

    def download(self):
        pass

    @property
    def raw_file_names(self):
        return ["test.pkl", "train.pkl", "validation.pkl"]

    def _perform_data_preparation(self, *args, **kwargs):
        print("Check for raw data in", self.raw_dir)
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            raise ValueError("Raw data is missing")

    @staticmethod
    def _load_tuples(input_file_path):
        with open(input_file_path, "rb") as input_file:
            for row in pickle.load(input_file).values:
                yield row[1], row[2:].astype(bool), row[0]

    @staticmethod
    def _get_data_size(input_file_path):
        with open(input_file_path, "rb") as f:
            return len(pickle.load(f))

    def setup_processed(self):
        print("Transform splits")
        os.makedirs(self.processed_dir, exist_ok=True)
        for k in ["test", "train", "validation"]:
            print("transform", k)
            torch.save(
                self._load_data_from_file(os.path.join(self.raw_dir, f"{k}.pkl")),
                os.path.join(self.processed_dir, f"{k}.pt"),
            )


class JCIData(JCIBase):
    READER = dr.OrdReader


class JCISelfies(JCIBase):
    READER = dr.SelfiesReader


class JCITokenData(JCIBase):
    READER = dr.ChemDataReader


class _ChEBIDataExtractor(_DynamicDataset, ABC):
    """
    A class for extracting and processing data from the ChEBI dataset.

    Args:
        chebi_version_train (int, optional): The version of ChEBI to use for training and validation. If not set,
            chebi_version will be used for training, validation and test. Defaults to None.
        single_class (int, optional): The ID of the single class to predict. If not set, all available labels will be
            predicted. Defaults to None.
        subset (Literal["2_STAR", "3_STAR"], optional): If set, only use entities that are part of the given subset.
        **kwargs: Additional keyword arguments (passed to XYBaseDataModule).

    Attributes:
        single_class (Optional[int]): The ID of the single class to predict.
        chebi_version_train (Optional[int]): The version of ChEBI to use for training and validation.
        subset (Optional[Literal["2_STAR", "3_STAR"]]): If set, only use entities that are part of the given subset.

    """

    # ---- Index for columns of processed `data.pkl` (derived from `_graph_to_raw_dataset` method) ------
    # "id" at                 row index 0
    # "name" at               row index 1
    # "SMILES" at             row index 2
    # labels starting from    row index 3
    _ID_IDX: int = 0
    _DATA_REPRESENTATION_IDX: int = 2
    _LABELS_START_IDX: int = 3

    def __init__(
        self,
        chebi_version_train: Optional[int] = None,
        single_class: Optional[int] = None,
        subset: Optional[Literal["2_STAR", "3_STAR"]] = None,
        augment_smiles: bool = False,
        aug_smiles_variations: Optional[int] = None,
        **kwargs,
    ):
        if bool(augment_smiles):
            assert (
                int(aug_smiles_variations) > 0
            ), "Number of variations must be greater than 0"
            aug_smiles_variations = int(aug_smiles_variations)

            if not kwargs.get("splits_file_path", None):
                raise ValueError(
                    "When using SMILES augmentation, a splits_file_path must be provided to ensure consistent splits."
                )

            reader_kwargs = kwargs.get("reader_kwargs", {})
            reader_kwargs["canonicalize_smiles"] = False
            kwargs["reader_kwargs"] = reader_kwargs

        self.augment_smiles = bool(augment_smiles)
        self.aug_smiles_variations = aug_smiles_variations
        # predict only single class (given as id of one of the classes present in the raw data set)
        self.single_class = single_class
        self.subset = subset

        super(_ChEBIDataExtractor, self).__init__(**kwargs)
        # use different version of chebi for training and validation (if not None)
        # (still uses self.chebi_version for test set)
        self.chebi_version_train = chebi_version_train

        if self.chebi_version_train is not None:
            # Instantiate another same class with "chebi_version" as "chebi_version_train", if train_version is given
            # This is to get the data from respective directory related to "chebi_version_train"
            _init_kwargs = kwargs
            _init_kwargs["chebi_version"] = self.chebi_version_train
            self._chebi_version_train_obj = self.__class__(
                single_class=self.single_class,
                augment_smiles=self.augment_smiles,
                aug_smiles_variations=self.aug_smiles_variations,
                **_init_kwargs,
            )

    # ------------------------------ Phase: Prepare data -----------------------------------
    def _perform_data_preparation(self, *args: Any, **kwargs: Any) -> None:
        """
        Prepares the data for the Chebi dataset.

        This method checks for the presence of raw data in the specified directory.
        If the raw data is missing, it fetches the ontology and creates a dataframe & saves it as data.pkl pickle file.

        The resulting dataframe/pickle file is expected to contain columns with the following structure:
            - Column at index `self._ID_IDX`: ID of chebi data instance
            - Column at index `self._DATA_REPRESENTATION_IDX`: SMILES representation of the chemical
            - Column from index `self._LABELS_START_IDX` onwards: Labels

        It will pre-process the data related to `chebi_version_train`, if specified.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None
        """
        super()._perform_data_preparation(args, kwargs)

        if self.chebi_version_train is not None:
            if not os.path.isfile(
                os.path.join(
                    self._chebi_version_train_obj.processed_dir_main,
                    self._chebi_version_train_obj.processed_main_file_names_dict[
                        "data"
                    ],
                )
            ):
                print(
                    f"Missing processed data related to train version: {self.chebi_version_train}"
                )
                print("Calling the prepare_data method related to it")
                # Generate the "chebi_version_train" data if it doesn't exist
                self._chebi_version_train_obj.prepare_data(*args, **kwargs)

    def _download_required_data(self) -> str:
        """
        Downloads the required raw data related to chebi.

        Returns:
            str: Path to the downloaded data.
        """
        return self._load_chebi(self.chebi_version)

    def _load_chebi(self, version: int) -> str:
        """
        Load the ChEBI ontology file.

        Args:
            version (int): The version of the ChEBI ontology to load.

        Returns:
            str: The file path of the loaded ChEBI ontology.
        """
        import requests

        chebi_name = self.raw_file_names_dict["chebi"]
        chebi_path = os.path.join(self.raw_dir, chebi_name)
        if not os.path.isfile(chebi_path):
            print(
                f"Missing raw chebi data related to version: v_{version}, Downloading..."
            )
            url = f"http://purl.obolibrary.org/obo/chebi/{version}/chebi.obo"
            r = requests.get(url, allow_redirects=True)
            open(chebi_path, "wb").write(r.content)
        return chebi_path

    def _extract_class_hierarchy(self, data_path: str) -> "nx.DiGraph":
        """
        Extracts the class hierarchy from the ChEBI ontology.
        Constructs a directed graph (DiGraph) using NetworkX, where nodes are annotated with fields/terms from
        the chebi term documents from `.obo` file.

        Args:
            data_path (str): The path to the ChEBI ontology.

        Returns:
            nx.DiGraph: The class hierarchy.
        """
        import fastobo
        import networkx as nx

        with open(data_path, encoding="utf-8") as chebi:
            chebi = "\n".join(line for line in chebi if not line.startswith("xref:"))

        elements = []
        for term_doc in fastobo.loads(chebi):
            if (
                term_doc
                and isinstance(term_doc.id, fastobo.id.PrefixedIdent)
                and term_doc.id.prefix == "CHEBI"
            ):
                term_dict = term_callback(term_doc)
                if term_dict and (
                    not self.subset or term_dict["subset"] == self.subset
                ):
                    elements.append(term_dict)

        g = nx.DiGraph()
        for n in elements:
            g.add_node(n["id"], **n)

        # Only take the edges which connects the existing nodes, to avoid internal creation of obsolete nodes
        # https://github.com/ChEB-AI/python-chebai/pull/55#issuecomment-2386654142
        g.add_edges_from(
            [(p, q["id"]) for q in elements for p in q["parents"] if g.has_node(p)]
        )

        print("Compute transitive closure")
        return nx.transitive_closure_dag(g)

    def _graph_to_raw_dataset(self, g: "nx.DiGraph") -> pd.DataFrame:
        """
        Converts the graph to a raw dataset.
        Uses the graph created by `_extract_class_hierarchy` method to extract the
        raw data in Dataframe format with additional columns corresponding to each multi-label class.

        Args:
            g (nx.DiGraph): The class hierarchy graph.

        Returns:
            pd.DataFrame: The raw dataset created from the graph.
        """
        import networkx as nx

        smiles = nx.get_node_attributes(g, "smiles")
        names = nx.get_node_attributes(g, "name")

        print("Process graph")

        molecules, smiles_list = zip(
            *(
                (n, smiles)
                for n, smiles in ((n, smiles.get(n)) for n in smiles.keys())
                if smiles
            )
        )
        data = OrderedDict(id=molecules)  # `id` column at index 0
        data["name"] = [
            names.get(node) for node in molecules
        ]  # `name` column at index 1
        data["SMILES"] = smiles_list  # `SMILES` (data representation) column at index 2

        # Labels columns from index 3 onwards
        for n in self.select_classes(g):
            data[n] = [
                ((n in g.predecessors(node)) or (n == node)) for node in molecules
            ]

        data = pd.DataFrame(data)
        data = data[~data["SMILES"].isnull()]
        data = data[~data["name"].isin(CHEBI_BLACKLIST)]

        return data

    def _after_prepare_data(self, *args, **kwargs) -> None:
        self._perform_smiles_augmentation()

    def _perform_smiles_augmentation(self) -> None:
        if not self.augment_smiles:
            return

        aug_pkl_file_name = self.processed_main_file_names_dict["aug_data"]
        aug_data_df = self.get_processed_pickled_df_file(aug_pkl_file_name)
        if aug_data_df is not None:
            self._data_pkl_filename = aug_pkl_file_name
            return

        data_df = self.get_processed_pickled_df_file(
            self.processed_main_file_names_dict["data"]
        )

        # +1 to account for if original SMILES is generated by random chance
        AUG_SMILES_VARIATIONS = self.aug_smiles_variations + 1

        def generate_augmented_smiles(smiles: str) -> list[str]:
            mol: Chem.Mol = Chem.MolFromSmiles(smiles)
            if mol is None:
                return [smiles]  # if mol is None, return original SMILES

            # sanitization set to False, as it can alter the fragment representation in ways you might not want.
            # As we don’t want RDKit to "fix" fragments, only need the fragments as-is, to generate SMILES strings.
            frags = Chem.GetMolFrags(mol, asMols=True, sanitizeFrags=False)
            augmented = set()

            frag_smiles: list[set] = []
            for frag in frags:
                atom_ids = [atom.GetIdx() for atom in frag.GetAtoms()]
                random.shuffle(atom_ids)  # seed set by lightning
                atom_id_iter = cycle(atom_ids)
                frag_smiles.append(
                    {
                        Chem.MolToSmiles(
                            frag, rootedAtAtom=next(atom_id_iter), doRandom=True
                        )
                        for _ in range(AUG_SMILES_VARIATIONS)
                    }
                )
            if len(frags) > 1:
                # all permutations (ignoring the set order, meaning mixing sets in every order),
                aug_counter: int = 0
                for perm in permutations(frag_smiles):
                    for combo in product(*perm):
                        augmented.add(".".join(combo))
                        aug_counter += 1
                        if aug_counter >= AUG_SMILES_VARIATIONS:
                            break
                    if aug_counter >= AUG_SMILES_VARIATIONS:
                        break
            else:
                augmented = frag_smiles[0]

            if smiles in augmented:
                augmented.remove(smiles)

            if len(augmented) > AUG_SMILES_VARIATIONS - 1:
                # AUG_SMILES_VARIATIONS = number of new smiles needed to generate + original smiles
                # if 3 smiles variations are needed, and 4 are generated because none
                # correponds to original smiles and no-one is elimnated in previous if condition
                augmented = random.sample(augmented, AUG_SMILES_VARIATIONS - 1)

            # original smiles always first in the list
            return [smiles] + list(augmented)

        data_df["SMILES"] = data_df["SMILES"].apply(generate_augmented_smiles)

        # Explode the list of augmented smiles into multiple rows
        # augmented smiles will have same ident, as of the original, but does it matter ?
        # instead its helpful to group augmented smiles generated from the same original SMILES
        exploded_df = data_df.explode("SMILES").reset_index(drop=True)
        self.save_processed(
            exploded_df, self.processed_main_file_names_dict["aug_data"]
        )
        self._data_pkl_filename = aug_pkl_file_name

    # ------------------------------ Phase: Setup data -----------------------------------
    def setup_processed(self) -> None:
        """
        Transform and prepare processed data for the ChEBI dataset.

        Main function of this method is to transform `data.pkl` into a model input data format (`data.pt`),
        ensuring that the data is in a format compatible for input to the model.
        The transformed data must contain the following keys: `ident`, `features`, `labels`, and `group`.
        This method uses a subclass of Data Reader to perform the transformation.

        It will transform the data related to `chebi_version_train`, if specified.
        """
        super().setup_processed()

        # Transform the data related to "chebi_version_train" to encoded data, if it doesn't exist
        if self.chebi_version_train is not None and not os.path.isfile(
            os.path.join(
                self._chebi_version_train_obj.processed_dir,
                self._chebi_version_train_obj.processed_file_names_dict["data"],
            )
        ):
            print(
                f"Missing encoded data related to train version: {self.chebi_version_train}"
            )
            print("Calling the setup method related to it")
            self._chebi_version_train_obj.setup()

    def _load_dict(self, input_file_path: str) -> Generator[dict[str, Any], None, None]:
        """
        Loads a dictionary from a pickled file, yielding individual dictionaries for each row.

        This method reads data from a specified pickled file, processes each row to extract relevant
        information, and yields dictionaries containing the keys `features`, `labels`, and `ident`.
        If `single_class` is specified, it only includes the label for that specific class; otherwise,
        it includes labels for all classes starting from the fourth column.

        The pickled file is expected to contain rows with the following structure:
            - Data at row index `self._ID_IDX`: ID of the chebi data instance
            - Data at row index `self._DATA_REPRESENTATION_IDX` : SMILES representation for the chemical
            - Data from row index `self._LABELS_START_IDX` onwards: Labels

        This method is used in `_load_data_from_file` to process each row of data and convert it
        into the desired dictionary format before loading it into the model.

        Args:
            input_file_path (str): The path to the input pickled file.

        Yields:
            Dict[str, Any]: A dictionary with keys `features`, `labels`, and `ident`.
            `features` contains the sequence, `labels` contains the labels as boolean values,
            and `ident` contains the identifier.
        """
        with open(input_file_path, "rb") as input_file:
            df = pd.read_pickle(input_file)

            if self.single_class is None:
                all_labels = df.iloc[:, self._LABELS_START_IDX :].to_numpy(dtype=bool)
            else:
                single_cls_index = df.columns.get_loc(int(self.single_class))
                all_labels = df.iloc[:, [single_cls_index]].to_numpy(dtype=bool)

            features = df.iloc[:, self._DATA_REPRESENTATION_IDX].to_numpy()
            idents = df.iloc[:, self._ID_IDX].to_numpy()

            for feat, labels, ident in zip(features, all_labels, idents):
                yield dict(features=feat, labels=labels, ident=ident)

    # ------------------------------ Phase: Dynamic Splits -----------------------------------
    def _get_data_splits(self) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Loads encoded/transformed data and generates training, validation, and test splits.

        This method first loads encoded data from a file named `data.pt`, which is derived from either
        `chebi_version` or `chebi_version_train`. It then splits the data into training, validation, and test sets.

        If `chebi_version_train` is provided:
            - Loads additional encoded data from `chebi_version_train`.
            - Splits this data into training and validation sets, while using the test set from `chebi_version`.
            - Prunes the test set from `chebi_version` to include only labels that exist in `chebi_version_train`.

        If `chebi_version_train` is not provided:
            - Splits the data from `chebi_version` into training, validation, and test sets without modification.

        Raises:
            FileNotFoundError: If the required `data.pt` file(s) do not exist. Ensure that `prepare_data`
            and/or `setup` methods have been called to generate the dataset files.

        Returns:
            Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]: A tuple containing three DataFrames:
                - Training set
                - Validation set
                - Test set
        """
        try:
            filename = self.processed_file_names_dict["data"]
            data_chebi_version = self.load_processed_data_from_file(
                os.path.join(self.processed_dir, filename)
            )
        except FileNotFoundError:
            raise FileNotFoundError(
                "File data.pt doesn't exists. "
                "Please call 'prepare_data' and/or 'setup' methods to generate the dataset files"
            )

        df_chebi_version = pd.DataFrame(data_chebi_version)
        train_df_chebi_ver, df_test_chebi_ver = self.get_test_split(
            df_chebi_version, seed=self.dynamic_data_split_seed
        )

        if self.chebi_version_train is not None:
            # Load encoded data derived from "chebi_version_train"
            try:
                filename_train = (
                    self._chebi_version_train_obj.processed_file_names_dict["data"]
                )
                data_chebi_train_version = torch.load(
                    os.path.join(
                        self._chebi_version_train_obj.processed_dir, filename_train
                    ),
                    weights_only=False,
                )
            except FileNotFoundError:
                raise FileNotFoundError(
                    f"File data.pt doesn't exists related to chebi_version_train {self.chebi_version_train}."
                    f"Please call 'prepare_data' and/or 'setup' methods to generate the dataset files"
                )

            df_chebi_train_version = pd.DataFrame(data_chebi_train_version)
            # Get train/val split of data based on "chebi_version_train", but
            # using test set from "chebi_version"
            df_train, df_val = self.get_train_val_splits_given_test(
                df_chebi_train_version,
                df_test_chebi_ver,
                seed=self.dynamic_data_split_seed,
            )
            # Modify test set from "chebi_version" to only include the labels that
            # exists in "chebi_version_train", all other entries remains same.
            df_test = self._setup_pruned_test_set(df_test_chebi_ver)
        else:
            # Get all splits based on "chebi_version"
            df_train, df_val = self.get_train_val_splits_given_test(
                train_df_chebi_ver,
                df_test_chebi_ver,
                seed=self.dynamic_data_split_seed,
            )
            df_test = df_test_chebi_ver

        return df_train, df_val, df_test

    def _setup_pruned_test_set(
        self, df_test_chebi_version: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Create a test set with the same leaf nodes, but use only classes that appear in the training set.

        Args:
            df_test_chebi_version (pd.DataFrame): The test dataset.

        Returns:
            pd.DataFrame: The pruned test dataset.
        """
        classes_file_name = "classes.txt"

        # Load original and new classes
        with open(os.path.join(self.processed_dir_main, classes_file_name), "r") as f:
            orig_classes = f.readlines()
        with open(
            os.path.join(
                self._chebi_version_train_obj.processed_dir_main, classes_file_name
            ),
            "r",
        ) as f:
            new_classes = f.readlines()

        # Mapping array (-1 means no match in new classes)
        mapping_array = np.array(
            [
                -1 if oc not in new_classes else new_classes.index(oc)
                for oc in orig_classes
            ],
            dtype=int,
        )

        # Convert labels column to 2D NumPy array
        labels_matrix = np.array(df_test_chebi_version["labels"].tolist(), dtype=bool)

        # Allocate new labels matrix
        num_new_classes = len(new_classes)
        new_labels_matrix = np.zeros(
            (labels_matrix.shape[0], num_new_classes), dtype=bool
        )

        # Copy only valid columns
        valid_mask = mapping_array != -1
        new_labels_matrix[:, mapping_array[valid_mask]] = labels_matrix[:, valid_mask]

        # Assign back
        df_test_chebi_version["labels"] = new_labels_matrix.tolist()
        return df_test_chebi_version

    # ------------------------------ Phase: Raw Properties -----------------------------------
    @property
    def base_dir(self) -> str:
        """
        Return the base directory path for data.

        Returns:
            str: The base directory path for data.
        """
        return os.path.join("data", f"chebi_v{self.chebi_version}")

    @property
    def processed_dir_main(self) -> str:
        """
        Returns the main directory path where processed data is stored.

        Returns:
            str: The path to the main processed data directory, based on the base directory and the instance's name.
        """
        return os.path.join(
            self.base_dir,
            self._name if self.subset is None else f"{self._name}_{self.subset}",
            "processed",
        )

    @property
    def processed_dir(self) -> str:
        """
        Return the directory path for processed data.

        Returns:
            str: The path to the processed data directory.
        """
        res = os.path.join(
            self.processed_dir_main,
            *self.identifier,
        )
        if self.single_class is None:
            return res
        else:
            return os.path.join(res, f"single_{self.single_class}")

    @property
    def raw_file_names_dict(self) -> dict:
        return {"chebi": "chebi.obo"}

    @property
    def processed_main_file_names_dict(self) -> dict:
        """
        Returns a dictionary mapping processed data file names.

        Returns:
            dict: A dictionary mapping dataset key to their respective file names.
                  For example, {"data": "data.pkl"}.
        """
        p_dict = super().processed_main_file_names_dict
        if self.augment_smiles:
            p_dict["aug_data"] = f"aug_data_var{self.aug_smiles_variations}.pkl"
        return p_dict

    @property
    def processed_file_names_dict(self) -> dict:
        """
        Returns a dictionary for the processed and tokenized data files.

        Returns:
            dict: A dictionary mapping dataset keys to their respective file names.
                  For example, {"data": "data.pt"}.
        """
        if not self.augment_smiles:
            return super().processed_file_names_dict
        if self.n_token_limit is not None:
            return {
                "data": f"aug_data_var{self.aug_smiles_variations}_maxlen{self.n_token_limit}.pt"
            }
        return {"data": f"aug_data_var{self.aug_smiles_variations}.pt"}


class JCIExtendedBase(_ChEBIDataExtractor):
    @property
    def _name(self):
        return "JCI_extended"

    def select_classes(self, g, *args, **kwargs):
        return JCI_500_COLUMNS_INT


class ChEBIOverX(_ChEBIDataExtractor):
    """
    A class for extracting data from the ChEBI dataset with a threshold for selecting classes.
    This class is designed to filter Chebi classes based on a specified threshold, selecting only those classes
    which have a certain number of subclasses in the hierarchy.

    Attributes:
        LABEL_INDEX (int): The index of the label in the dataset.
        SMILES_INDEX (int): The index of the SMILES string in the dataset.
        READER (ChemDataReader): The reader used for reading the dataset.
        THRESHOLD (None): The threshold for selecting classes.
    """

    READER: dr.ChemDataReader = dr.ChemDataReader
    THRESHOLD: int = None

    @property
    def _name(self) -> str:
        """
        Returns the name of the dataset.

        Returns:
            str: The dataset name.
        """
        return f"ChEBI{self.THRESHOLD}"

    def select_classes(self, g: "nx.DiGraph", *args, **kwargs) -> List:
        """
        Selects classes from the ChEBI dataset based on the number of successors meeting a specified threshold.

        This method iterates over the nodes in the graph, counting the number of successors for each node.
        Nodes with a number of successors greater than or equal to the defined threshold are selected.

        Note:
            The input graph must be transitive closure of a directed acyclic graph.

        Args:
            g (nx.Graph): The graph representing the dataset.
            *args: Additional positional arguments (not used).
            **kwargs: Additional keyword arguments (not used).

        Returns:
            List: A sorted list of node IDs that meet the successor threshold criteria.

        Side Effects:
            Writes the list of selected nodes to a file named "classes.txt" in the specified processed directory.

        Notes:
            - The `THRESHOLD` attribute should be defined in the subclass of this class.
            - Nodes without a 'smiles' attribute are ignored in the successor count.
        """
        import networkx as nx

        smiles = nx.get_node_attributes(g, "smiles")
        nodes = list(
            sorted(
                {
                    node
                    for node in g.nodes
                    if sum(
                        1 if smiles[s] is not None else 0 for s in g.successors(node)
                    )
                    >= self.THRESHOLD
                }
            )
        )
        filename = "classes.txt"
        with open(os.path.join(self.processed_dir_main, filename), "wt") as fout:
            fout.writelines(str(node) + "\n" for node in nodes)
        return nodes


class ChEBIOverXDeepSMILES(ChEBIOverX):
    """
    A class for extracting data from the ChEBI dataset with DeepChem SMILES reader.

    Inherits from ChEBIOverX.

    Attributes:
        READER (DeepChemDataReader): The reader used for reading the dataset (DeepChem SMILES).
    """

    READER: dr.DeepChemDataReader = dr.DeepChemDataReader


class ChEBIOverXSELFIES(ChEBIOverX):
    """
    A class for extracting data from the ChEBI dataset with SELFIES reader.

    Inherits from ChEBIOverX.

    Attributes:
        READER (SelfiesReader): The reader used for reading the dataset (SELFIES).
    """

    READER: dr.SelfiesReader = dr.SelfiesReader


class ChEBIOver100(ChEBIOverX):
    """
    A class for extracting data from the ChEBI dataset with a threshold of 100 for selecting classes.

    Inherits from ChEBIOverX.

    Attributes:
        THRESHOLD (int): The threshold for selecting classes (100).
    """

    THRESHOLD: int = 100


class ChEBIOver50(ChEBIOverX):
    """
    A class for extracting data from the ChEBI dataset with a threshold of 50 for selecting classes.

    Inherits from ChEBIOverX.

    Attributes:
        THRESHOLD (int): The threshold for selecting classes (50).
    """

    THRESHOLD: int = 50


class ChEBIOver100DeepSMILES(ChEBIOverXDeepSMILES, ChEBIOver100):
    """
    A class for extracting data from the ChEBI dataset with DeepChem SMILES reader and a threshold of 100.

    Inherits from ChEBIOverXDeepSMILES and ChEBIOver100.
    """

    pass


class ChEBIOver100SELFIES(ChEBIOverXSELFIES, ChEBIOver100):
    """
    A class for extracting data from the ChEBI dataset with SELFIES reader and a threshold of 100.

    Inherits from ChEBIOverXSELFIES and ChEBIOver100.
    """

    pass


class ChEBIOver50SELFIES(ChEBIOverXSELFIES, ChEBIOver50):
    pass


class ChEBIOverXPartial(ChEBIOverX):
    """
    Dataset that doesn't use the full ChEBI, but extracts a part of ChEBI (subclasses of a given top class)

    Attributes:
        top_class_id (int): The ID of the top class from which to extract subclasses.
    """

    def __init__(self, top_class_id: int, external_data_ratio: float, **kwargs):
        """
        Initializes the ChEBIOverXPartial dataset.

        Args:
            top_class_id (int): The ID of the top class from which to extract subclasses.
            **kwargs: Additional keyword arguments passed to the superclass initializer.
            external_data_ratio (float): How much external data (i.e., samples where top_class_id
            is no positive label) to include in the dataset. 0 means no external data, 1 means
            the maximum amount (i.e., the complete ChEBI dataset).
        """
        if "top_class_id" not in kwargs:
            kwargs["top_class_id"] = top_class_id
        if "external_data_ratio" not in kwargs:
            kwargs["external_data_ratio"] = external_data_ratio

        self.top_class_id: int = top_class_id
        self.external_data_ratio: float = external_data_ratio
        super().__init__(**kwargs)

    @property
    def processed_dir_main(self) -> str:
        """
        Returns the main processed data directory specific to the top class.

        Returns:
            str: The processed data directory path.
        """
        return os.path.join(
            self.base_dir,
            self._name,
            f"partial_{self.top_class_id}_ext_ratio_{self.external_data_ratio:.2f}",
            "processed",
        )

    def _extract_class_hierarchy(self, chebi_path: str) -> "nx.DiGraph":
        """
        Extracts a subset of ChEBI based on subclasses of the top class ID.

        This method calls the superclass method to extract the full class hierarchy,
        then extracts the subgraph containing only the descendants of the top class ID, including itself.

        Args:
            chebi_path (str): The file path to the ChEBI ontology file.

        Returns:
            nx.DiGraph: The extracted class hierarchy as a directed graph, limited to the
            descendants of the top class ID.
        """
        g = super()._extract_class_hierarchy(chebi_path)
        top_class_successors = list(g.successors(self.top_class_id)) + [
            self.top_class_id
        ]
        external_nodes = list(set(n for n in g.nodes if n not in top_class_successors))
        if 0 < self.external_data_ratio < 1:
            n_external_nodes = int(
                len(top_class_successors)
                * self.external_data_ratio
                / (1 - self.external_data_ratio)
            )
            print(
                f"Extracting {n_external_nodes} external nodes from the ChEBI dataset (ratio: {self.external_data_ratio:.2f})"
            )
            external_nodes = external_nodes[: int(n_external_nodes)]
        elif self.external_data_ratio == 0:
            external_nodes = []

        g = g.subgraph(top_class_successors + external_nodes)
        print(
            f"Subgraph contains {len(g.nodes)} nodes, of which {len(top_class_successors)} are subclasses of the top class ID {self.top_class_id}."
        )
        return g

    def select_classes(self, g: "nx.DiGraph", *args, **kwargs) -> List:
        """Only selects classes that meet the threshold AND are subclasses of the top class ID (including itself)."""
        import networkx as nx

        smiles = nx.get_node_attributes(g, "smiles")
        nodes = list(
            sorted(
                {
                    node
                    for node in g.nodes
                    if sum(
                        1 if smiles[s] is not None else 0 for s in g.successors(node)
                    )
                    >= self.THRESHOLD
                    and (
                        self.top_class_id in g.predecessors(node)
                        or node == self.top_class_id
                    )
                }
            )
        )
        filename = "classes.txt"
        with open(os.path.join(self.processed_dir_main, filename), "wt") as fout:
            fout.writelines(str(node) + "\n" for node in nodes)
        return nodes


class ChEBIOver50Partial(ChEBIOverXPartial, ChEBIOver50):
    """
    Dataset that extracts a part of ChEBI based on subclasses of a given top class,
    with a threshold of 50 for selecting classes.

    Inherits from ChEBIOverXPartial and ChEBIOver50.
    """

    pass


class ChEBIOverXFingerprints(ChEBIOverX):
    """A class that uses Fingerprints for the processed data (used for fixed-length ML models)."""

    READER = dr.FingerprintReader


class ChEBIOver100Fingerprints(ChEBIOverXFingerprints, ChEBIOver100):
    """
    A class for extracting data from the ChEBI dataset with Fingerprints reader and a threshold of 100.

    Inherits from ChEBIOverXFingerprints and ChEBIOver100.
    """

    pass


class JCIExtendedBPEData(JCIExtendedBase):
    READER = dr.ChemBPEReader


class JCIExtendedTokenData(JCIExtendedBase):
    READER = dr.ChemDataReader


class JCIExtSelfies(JCIExtendedBase):
    READER = dr.SelfiesReader


def chebi_to_int(s: str) -> int:
    """
    Converts a ChEBI term string representation to an integer ID.

    Args:
    - s (str): A ChEBI term string, e.g., "CHEBI:12345".

    Returns:
    - int: The integer ID extracted from the ChEBI term string.
    """
    return int(s[s.index(":") + 1 :])


def term_callback(doc: "fastobo.term.TermFrame") -> Union[Dict, bool]:
    """
    Extracts information from a ChEBI term document.
    This function takes a ChEBI term document as input and extracts relevant information such as the term ID, parents,
    parts, name, and SMILES string. It returns a dictionary containing the extracted information.

    Args:
    - doc: A ChEBI term document.

    Returns:
    A dictionary containing the following keys:
    - "id": The ID of the ChEBI term.
    - "parents": A list of parent term IDs.
    - "has_part": A set of term IDs representing the parts of the ChEBI term.
    - "name": The name of the ChEBI term.
    - "smiles": The SMILES string associated with the ChEBI term, if available.
    """
    import fastobo

    parts = set()
    parents = []
    name = None
    smiles = None
    subset = None
    for clause in doc:
        if isinstance(clause, fastobo.term.PropertyValueClause):
            t = clause.property_value
            if str(t.relation) == "http://purl.obolibrary.org/obo/chebi/smiles":
                assert smiles is None
                smiles = t.value
        # in older chebi versions, smiles strings are synonyms
        # e.g. synonym: "[F-].[Na+]" RELATED SMILES [ChEBI]
        elif isinstance(clause, fastobo.term.SynonymClause):
            if "SMILES" in clause.raw_value():
                assert smiles is None
                smiles = clause.raw_value().split('"')[1]
        elif isinstance(clause, fastobo.term.RelationshipClause):
            if str(clause.typedef) == "has_part":
                parts.add(chebi_to_int(str(clause.term)))
        elif isinstance(clause, fastobo.term.IsAClause):
            parents.append(chebi_to_int(str(clause.term)))
        elif isinstance(clause, fastobo.term.NameClause):
            name = str(clause.name)
        elif isinstance(clause, fastobo.term.SubsetClause):
            subset = str(clause.subset)

        if isinstance(clause, fastobo.term.IsObsoleteClause):
            if clause.obsolete:
                # if the term document contains clause as obsolete as true, skips this document.
                return False

    return {
        "id": chebi_to_int(str(doc.id)),
        "parents": parents,
        "has_part": parts,
        "name": name,
        "smiles": smiles,
        "subset": subset,
    }


atom_index = (
    r"\*",
    "H",
    "He",
    "Li",
    "Be",
    "B",
    "C",
    "N",
    "O",
    "F",
    "Ne",
    "Na",
    "Mg",
    "Al",
    "Si",
    "P",
    "S",
    "Cl",
    "Ar",
    "K",
    "Ca",
    "Sc",
    "Ti",
    "V",
    "Cr",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Zn",
    "Ga",
    "Ge",
    "As",
    "Se",
    "Br",
    "Kr",
    "Rb",
    "Sr",
    "Y",
    "Zr",
    "Nb",
    "Mo",
    "Tc",
    "Ru",
    "Rh",
    "Pd",
    "Ag",
    "Cd",
    "In",
    "Sn",
    "Sb",
    "Te",
    "I",
    "Xe",
    "Cs",
    "Ba",
    "La",
    "Hf",
    "Ta",
    "W",
    "Re",
    "Os",
    "Ir",
    "Pt",
    "Au",
    "Hg",
    "Tl",
    "Pb",
    "Bi",
    "Po",
    "At",
    "Rn",
    "Fr",
    "Ra",
    "Ac",
    "Rf",
    "Db",
    "Sg",
    "Bh",
    "Hs",
    "Mt",
    "Ds",
    "Rg",
    "Cn",
    "Nh",
    "Fl",
    "Mc",
    "Lv",
    "Ts",
    "Og",
    "Ce",
    "Pr",
    "Nd",
    "Pm",
    "Sm",
    "Eu",
    "Gd",
    "Tb",
    "Dy",
    "Ho",
    "Er",
    "Tm",
    "Yb",
    "Lu",
    "Th",
    "Pa",
    "U",
    "Np",
    "Pu",
    "Am",
    "Cm",
    "Bk",
    "Cf",
    "Es",
    "Fm",
    "Md",
    "No",
    "Lr",
    "c",
    "n",
    "s",
    "o",
    "se",
    "p",
)

JCI_500_COLUMNS = [
    "CHEBI:25716",
    "CHEBI:72010",
    "CHEBI:60926",
    "CHEBI:39206",
    "CHEBI:24315",
    "CHEBI:22693",
    "CHEBI:23981",
    "CHEBI:23066",
    "CHEBI:35343",
    "CHEBI:18303",
    "CHEBI:60971",
    "CHEBI:35753",
    "CHEBI:24836",
    "CHEBI:59268",
    "CHEBI:35992",
    "CHEBI:51718",
    "CHEBI:27093",
    "CHEBI:38311",
    "CHEBI:46940",
    "CHEBI:26399",
    "CHEBI:27325",
    "CHEBI:33637",
    "CHEBI:37010",
    "CHEBI:36786",
    "CHEBI:59777",
    "CHEBI:36871",
    "CHEBI:26799",
    "CHEBI:50525",
    "CHEBI:26848",
    "CHEBI:52782",
    "CHEBI:75885",
    "CHEBI:37533",
    "CHEBI:47018",
    "CHEBI:27150",
    "CHEBI:26707",
    "CHEBI:131871",
    "CHEBI:134179",
    "CHEBI:24727",
    "CHEBI:59238",
    "CHEBI:26373",
    "CHEBI:46774",
    "CHEBI:33642",
    "CHEBI:38686",
    "CHEBI:74222",
    "CHEBI:23666",
    "CHEBI:46770",
    "CHEBI:16460",
    "CHEBI:37485",
    "CHEBI:21644",
    "CHEBI:52565",
    "CHEBI:33576",
    "CHEBI:76170",
    "CHEBI:46640",
    "CHEBI:61902",
    "CHEBI:22750",
    "CHEBI:35348",
    "CHEBI:48030",
    "CHEBI:2571",
    "CHEBI:38131",
    "CHEBI:83575",
    "CHEBI:136889",
    "CHEBI:26250",
    "CHEBI:36244",
    "CHEBI:23906",
    "CHEBI:38261",
    "CHEBI:22916",
    "CHEBI:35924",
    "CHEBI:24689",
    "CHEBI:32877",
    "CHEBI:50511",
    "CHEBI:26588",
    "CHEBI:24385",
    "CHEBI:5653",
    "CHEBI:48591",
    "CHEBI:38295",
    "CHEBI:58944",
    "CHEBI:134396",
    "CHEBI:49172",
    "CHEBI:26558",
    "CHEBI:64708",
    "CHEBI:35923",
    "CHEBI:25961",
    "CHEBI:47779",
    "CHEBI:46812",
    "CHEBI:37863",
    "CHEBI:22718",
    "CHEBI:36562",
    "CHEBI:38771",
    "CHEBI:36078",
    "CHEBI:26935",
    "CHEBI:33555",
    "CHEBI:23044",
    "CHEBI:15693",
    "CHEBI:33892",
    "CHEBI:33909",
    "CHEBI:35766",
    "CHEBI:51149",
    "CHEBI:35972",
    "CHEBI:38304",
    "CHEBI:46942",
    "CHEBI:24026",
    "CHEBI:33721",
    "CHEBI:38093",
    "CHEBI:38830",
    "CHEBI:26875",
    "CHEBI:37963",
    "CHEBI:61910",
    "CHEBI:47891",
    "CHEBI:74818",
    "CHEBI:50401",
    "CHEBI:24834",
    "CHEBI:33299",
    "CHEBI:63424",
    "CHEBI:63427",
    "CHEBI:15841",
    "CHEBI:33666",
    "CHEBI:26214",
    "CHEBI:22484",
    "CHEBI:27024",
    "CHEBI:46845",
    "CHEBI:64365",
    "CHEBI:63566",
    "CHEBI:38757",
    "CHEBI:83264",
    "CHEBI:24867",
    "CHEBI:37841",
    "CHEBI:33720",
    "CHEBI:36885",
    "CHEBI:59412",
    "CHEBI:64612",
    "CHEBI:36500",
    "CHEBI:37015",
    "CHEBI:84135",
    "CHEBI:51751",
    "CHEBI:18133",
    "CHEBI:57613",
    "CHEBI:38976",
    "CHEBI:25810",
    "CHEBI:24873",
    "CHEBI:35571",
    "CHEBI:83812",
    "CHEBI:37909",
    "CHEBI:51750",
    "CHEBI:15889",
    "CHEBI:48470",
    "CHEBI:24676",
    "CHEBI:22480",
    "CHEBI:139051",
    "CHEBI:23252",
    "CHEBI:51454",
    "CHEBI:88061",
    "CHEBI:46874",
    "CHEBI:38338",
    "CHEBI:62618",
    "CHEBI:59266",
    "CHEBI:84403",
    "CHEBI:27116",
    "CHEBI:77632",
    "CHEBI:38418",
    "CHEBI:35213",
    "CHEBI:35496",
    "CHEBI:78799",
    "CHEBI:38314",
    "CHEBI:35568",
    "CHEBI:35573",
    "CHEBI:33847",
    "CHEBI:16038",
    "CHEBI:33741",
    "CHEBI:33654",
    "CHEBI:17387",
    "CHEBI:33572",
    "CHEBI:36233",
    "CHEBI:22297",
    "CHEBI:23990",
    "CHEBI:38102",
    "CHEBI:24436",
    "CHEBI:35189",
    "CHEBI:79202",
    "CHEBI:68489",
    "CHEBI:18254",
    "CHEBI:78189",
    "CHEBI:47019",
    "CHEBI:61655",
    "CHEBI:24373",
    "CHEBI:26347",
    "CHEBI:36709",
    "CHEBI:73539",
    "CHEBI:35507",
    "CHEBI:35293",
    "CHEBI:140326",
    "CHEBI:46668",
    "CHEBI:17188",
    "CHEBI:61109",
    "CHEBI:35819",
    "CHEBI:33744",
    "CHEBI:73474",
    "CHEBI:134361",
    "CHEBI:33238",
    "CHEBI:26766",
    "CHEBI:17517",
    "CHEBI:25508",
    "CHEBI:22580",
    "CHEBI:26394",
    "CHEBI:35356",
    "CHEBI:50918",
    "CHEBI:24860",
    "CHEBI:2468",
    "CHEBI:33581",
    "CHEBI:26519",
    "CHEBI:37948",
    "CHEBI:33823",
    "CHEBI:59554",
    "CHEBI:46848",
    "CHEBI:24897",
    "CHEBI:26893",
    "CHEBI:63394",
    "CHEBI:29348",
    "CHEBI:35790",
    "CHEBI:25241",
    "CHEBI:58958",
    "CHEBI:24397",
    "CHEBI:25413",
    "CHEBI:24302",
    "CHEBI:46850",
    "CHEBI:51867",
    "CHEBI:35314",
    "CHEBI:50893",
    "CHEBI:36130",
    "CHEBI:33558",
    "CHEBI:24782",
    "CHEBI:36087",
    "CHEBI:26649",
    "CHEBI:47923",
    "CHEBI:33184",
    "CHEBI:23643",
    "CHEBI:25985",
    "CHEBI:33257",
    "CHEBI:61355",
    "CHEBI:24697",
    "CHEBI:36838",
    "CHEBI:23451",
    "CHEBI:33242",
    "CHEBI:26872",
    "CHEBI:50523",
    "CHEBI:16701",
    "CHEBI:36699",
    "CHEBI:35505",
    "CHEBI:24360",
    "CHEBI:59737",
    "CHEBI:26455",
    "CHEBI:51285",
    "CHEBI:35504",
    "CHEBI:36309",
    "CHEBI:33554",
    "CHEBI:47909",
    "CHEBI:50858",
    "CHEBI:53339",
    "CHEBI:25609",
    "CHEBI:23665",
    "CHEBI:35902",
    "CHEBI:35552",
    "CHEBI:139592",
    "CHEBI:35724",
    "CHEBI:38337",
    "CHEBI:35241",
    "CHEBI:29075",
    "CHEBI:62941",
    "CHEBI:140345",
    "CHEBI:59769",
    "CHEBI:28863",
    "CHEBI:47882",
    "CHEBI:35903",
    "CHEBI:33641",
    "CHEBI:47784",
    "CHEBI:23079",
    "CHEBI:25036",
    "CHEBI:50018",
    "CHEBI:28874",
    "CHEBI:35276",
    "CHEBI:26764",
    "CHEBI:65323",
    "CHEBI:51276",
    "CHEBI:37022",
    "CHEBI:22478",
    "CHEBI:23449",
    "CHEBI:72823",
    "CHEBI:63567",
    "CHEBI:50753",
    "CHEBI:38785",
    "CHEBI:46952",
    "CHEBI:36914",
    "CHEBI:33653",
    "CHEBI:62937",
    "CHEBI:36315",
    "CHEBI:37667",
    "CHEBI:38835",
    "CHEBI:35315",
    "CHEBI:33551",
    "CHEBI:18154",
    "CHEBI:79346",
    "CHEBI:26932",
    "CHEBI:39203",
    "CHEBI:25235",
    "CHEBI:23003",
    "CHEBI:64583",
    "CHEBI:46955",
    "CHEBI:33658",
    "CHEBI:59202",
    "CHEBI:28892",
    "CHEBI:33599",
    "CHEBI:33259",
    "CHEBI:64611",
    "CHEBI:37947",
    "CHEBI:65321",
    "CHEBI:63571",
    "CHEBI:25830",
    "CHEBI:50492",
    "CHEBI:26961",
    "CHEBI:33482",
    "CHEBI:63436",
    "CHEBI:47017",
    "CHEBI:51681",
    "CHEBI:48901",
    "CHEBI:52575",
    "CHEBI:35683",
    "CHEBI:24353",
    "CHEBI:61778",
    "CHEBI:13248",
    "CHEBI:35990",
    "CHEBI:33485",
    "CHEBI:35871",
    "CHEBI:27933",
    "CHEBI:27136",
    "CHEBI:26407",
    "CHEBI:33566",
    "CHEBI:47880",
    "CHEBI:24921",
    "CHEBI:38077",
    "CHEBI:48975",
    "CHEBI:59835",
    "CHEBI:83273",
    "CHEBI:22562",
    "CHEBI:33838",
    "CHEBI:35627",
    "CHEBI:51614",
    "CHEBI:36836",
    "CHEBI:63423",
    "CHEBI:22331",
    "CHEBI:25529",
    "CHEBI:36314",
    "CHEBI:83822",
    "CHEBI:38164",
    "CHEBI:51006",
    "CHEBI:28965",
    "CHEBI:38716",
    "CHEBI:76567",
    "CHEBI:35381",
    "CHEBI:51269",
    "CHEBI:37141",
    "CHEBI:25872",
    "CHEBI:36526",
    "CHEBI:51702",
    "CHEBI:25106",
    "CHEBI:51737",
    "CHEBI:38672",
    "CHEBI:36132",
    "CHEBI:38700",
    "CHEBI:25558",
    "CHEBI:17855",
    "CHEBI:18946",
    "CHEBI:83565",
    "CHEBI:15705",
    "CHEBI:35186",
    "CHEBI:33694",
    "CHEBI:36711",
    "CHEBI:23403",
    "CHEBI:35238",
    "CHEBI:36807",
    "CHEBI:47788",
    "CHEBI:24531",
    "CHEBI:33663",
    "CHEBI:22715",
    "CHEBI:57560",
    "CHEBI:38163",
    "CHEBI:23899",
    "CHEBI:50994",
    "CHEBI:26776",
    "CHEBI:51569",
    "CHEBI:35259",
    "CHEBI:77636",
    "CHEBI:35727",
    "CHEBI:35786",
    "CHEBI:24780",
    "CHEBI:26714",
    "CHEBI:26712",
    "CHEBI:26819",
    "CHEBI:63944",
    "CHEBI:36520",
    "CHEBI:25409",
    "CHEBI:22928",
    "CHEBI:23824",
    "CHEBI:79020",
    "CHEBI:26605",
    "CHEBI:139588",
    "CHEBI:52396",
    "CHEBI:37668",
    "CHEBI:50995",
    "CHEBI:52395",
    "CHEBI:61777",
    "CHEBI:38445",
    "CHEBI:24698",
    "CHEBI:63551",
    "CHEBI:35693",
    "CHEBI:83403",
    "CHEBI:36094",
    "CHEBI:35479",
    "CHEBI:25704",
    "CHEBI:25754",
    "CHEBI:38958",
    "CHEBI:21731",
    "CHEBI:23697",
    "CHEBI:38260",
    "CHEBI:33861",
    "CHEBI:22485",
    "CHEBI:2580",
    "CHEBI:18379",
    "CHEBI:23424",
    "CHEBI:33296",
    "CHEBI:37554",
    "CHEBI:33839",
    "CHEBI:36054",
    "CHEBI:23232",
    "CHEBI:18035",
    "CHEBI:63353",
    "CHEBI:23114",
    "CHEBI:76578",
    "CHEBI:26208",
    "CHEBI:32955",
    "CHEBI:24922",
    "CHEBI:36141",
    "CHEBI:24043",
    "CHEBI:35692",
    "CHEBI:46867",
    "CHEBI:38530",
    "CHEBI:24654",
    "CHEBI:38032",
    "CHEBI:26820",
    "CHEBI:35789",
    "CHEBI:62732",
    "CHEBI:26912",
    "CHEBI:22160",
    "CHEBI:26410",
    "CHEBI:36059",
    "CHEBI:51069",
    "CHEBI:33570",
    "CHEBI:24129",
    "CHEBI:37826",
    "CHEBI:16385",
    "CHEBI:26822",
    "CHEBI:46761",
    "CHEBI:83925",
    "CHEBI:25248",
    "CHEBI:37581",
    "CHEBI:35748",
    "CHEBI:26195",
    "CHEBI:33958",
    "CHEBI:58342",
    "CHEBI:17478",
    "CHEBI:36834",
    "CHEBI:25513",
    "CHEBI:57643",
    "CHEBI:38298",
    "CHEBI:64482",
    "CHEBI:33240",
    "CHEBI:47622",
    "CHEBI:33704",
    "CHEBI:83820",
    "CHEBI:33676",
    "CHEBI:32952",
    "CHEBI:131927",
    "CHEBI:26188",
    "CHEBI:35716",
    "CHEBI:28963",
    "CHEBI:22798",
    "CHEBI:60980",
    "CHEBI:17984",
    "CHEBI:37240",
    "CHEBI:28868",
    "CHEBI:27208",
    "CHEBI:15904",
    "CHEBI:35715",
    "CHEBI:22251",
    "CHEBI:61078",
    "CHEBI:61079",
    "CHEBI:58946",
    "CHEBI:37123",
    "CHEBI:33497",
    "CHEBI:50699",
    "CHEBI:22475",
    "CHEBI:35436",
]

JCI_500_COLUMNS_INT = [int(n.split(":")[-1]) for n in JCI_500_COLUMNS]

if __name__ == "__main__":
    data_module_05 = ChEBIOver50Partial(
        chebi_version=241,
        splits_file_path=os.path.join(
            "data", "chebi_v241", "ChEBI50", "splits_80_10_10.csv"
        ),
        top_class_id=22712,
        external_data_ratio=0.5,
    )
    data_module_05.prepare_data()
    data_module_05.setup()



import os

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import torch
import tqdm

import chebai.models.electra as electra
from chebai.loss.pretraining import ElectraPreLoss
from chebai.result.base import ResultProcessor


def visualise_loss(logs_path):
    df = pd.read_csv(os.path.join(logs_path, "metrics.csv"))
    df_loss = df.melt(
        id_vars="epoch", value_vars=["val_loss_epoch", "train_loss_epoch"]
    )
    lineplt = sns.lineplot(df_loss, x="epoch", y="value", hue="variable")  # noqa: F841
    plt.savefig(os.path.join(logs_path, "f1_plot.png"))
    plt.show()


# get predictions from model
def evaluate_model(logs_base_path, model_filename, data_module):
    model = electra.ElectraPre.load_from_checkpoint(
        os.path.join(
            logs_base_path,
            "best_epoch=85_val_loss=0.0147_val_micro-f1=0.90.ckpt",
            model_filename,
        )
    )
    assert isinstance(model, electra.ElectraPre)
    collate = data_module.reader.COLLATOR()
    test_file = "test.pt"
    data_path = os.path.join(data_module.processed_dir, test_file)
    data_list = torch.load(data_path, weights_only=False)
    preds_list = []
    labels_list = []

    for row in tqdm.tqdm(data_list):
        processable_data = model._process_batch(collate([row]), 0)
        model_output = model(processable_data, **processable_data["model_kwargs"])
        preds, labels = model._get_prediction_and_labels(
            processable_data, processable_data["labels"], model_output
        )
        preds_list.append(preds)
        labels_list.append(labels)

    test_preds = torch.cat(preds_list)
    test_labels = torch.cat(labels_list)
    print(test_preds.shape)
    print(test_labels.shape)
    test_loss = ElectraPreLoss()
    print(f"Loss on test set: {test_loss(test_preds, test_labels)}")
    # f1_macro = MultilabelF1Score(test_preds.shape[1], average='macro')
    # f1_micro = MultilabelF1Score(test_preds.shape[1], average='micro')
    # print(f'Macro-F1 on test set with {test_preds.shape[1]} classes: {f1_macro(test_preds, test_labels):3f}')
    # print(f'Micro-F1 on test set with {test_preds.shape[1]} classes: {f1_micro(test_preds, test_labels):3f}')


class PretrainingResultProcessor(ResultProcessor):
    @classmethod
    def _identifier(cls) -> str:
        return "PretrainingResultProcessor"



import logging

from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from chebai.models.base import ChebaiBaseNet

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)


class ChemLSTM(ChebaiBaseNet):
    def __init__(
        self,
        out_d,
        in_d,
        num_classes,
        criterion: nn.Module = None,
        num_layers=6,
        dropout=0.2,
        **kwargs,
    ):
        super().__init__(
            out_dim=out_d,
            input_dim=in_d,
            criterion=criterion,
            num_classes=num_classes,
            **kwargs,
        )
        self.lstm = nn.LSTM(
            in_d,
            out_d,
            batch_first=True,
            dropout=dropout,
            bidirectional=True,
            num_layers=num_layers,
        )
        self.embedding = nn.Embedding(1400, in_d)
        self.output = nn.Sequential(
            nn.Linear(out_d * 2, out_d),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(out_d, num_classes),
        )

    def forward(self, data, *args, **kwargs):
        x = data["features"]
        x_lens = data["model_kwargs"]["lens"]
        x = self.embedding(x)
        x = pack_padded_sequence(x, x_lens, batch_first=True, enforce_sorted=False)
        x = self.lstm(x)[0]
        x = pad_packed_sequence(x, batch_first=True)[0][
            :, 0
        ]  # reduce sequence dimension to first element
        x = self.output(x)
        return x



from typing import Tuple

import numpy as np
import torch
from jsonargparse import CLI
from torchmetrics.functional.classification import multilabel_auroc

from chebai.callbacks.epoch_metrics import MacroF1
from chebai.result.utils import load_results_from_buffer


class EvaluatePredictions:
    def __init__(self, eval_dir: str):
        """
        Initializes the EvaluatePredictions class.

        Args:
            eval_dir (str): Path to the directory containing evaluation files.
        """
        self.eval_dir = eval_dir
        self.metrics = []
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.num_labels = None

    @staticmethod
    def validate_eval_dir(label_files: torch.Tensor, pred_files: torch.Tensor) -> None:
        """
        Validates that the number of labels matches the number of predictions,
        ensuring that they have the same shape.

        Args:
            label_files (torch.Tensor): Tensor containing label data.
            pred_files (torch.Tensor): Tensor containing prediction data.

        Raises:
            ValueError: If label and prediction tensors are mismatched in shape.
        """
        if label_files is None or pred_files is None:
            raise ValueError("Both label and prediction tensors must be provided.")

        # Check if the number of labels matches the number of predictions
        if label_files.shape[0] != pred_files.shape[0]:
            raise ValueError(
                "Number of label tensors does not match the number of prediction tensors."
            )

        # Validate that the last dimension matches the expected number of classes
        if label_files.shape[1] != pred_files.shape[1]:
            raise ValueError(
                "Label and prediction tensors must have the same shape in terms of class outputs."
            )

    def evaluate(self) -> None:
        """
        Loads predictions and labels, validates file correspondence, and calculates Multilabel AUROC and Fmax.
        """
        test_preds, test_labels = load_results_from_buffer(self.eval_dir, self.device)
        self.validate_eval_dir(test_labels, test_preds)
        self.num_labels = test_preds.shape[1]

        ml_auroc = multilabel_auroc(
            test_preds, test_labels, num_labels=self.num_labels
        ).item()

        print("Multilabel AUC-ROC:", ml_auroc)

        fmax, threshold = self.calculate_fmax(test_preds, test_labels)
        print(f"F-max : {fmax}, threshold: {threshold}")

    def calculate_fmax(
        self, test_preds: torch.Tensor, test_labels: torch.Tensor
    ) -> Tuple[float, float]:
        """
        Calculates the Fmax metric using the F1 score at various thresholds.

        Args:
            test_preds (torch.Tensor): Predicted scores for the labels.
            test_labels (torch.Tensor): True labels for the evaluation.

        Returns:
            Tuple[float, float]: The maximum F1 score and the corresponding threshold.
        """
        # https://github.com/bio-ontology-research-group/deepgo2/blob/main/deepgo/metrics.py#L51-L52
        thresholds = np.linspace(0, 1, 101)
        fmax = 0.0
        best_threshold = 0.0

        for t in thresholds:
            custom_f1_metric = MacroF1(num_labels=self.num_labels, threshold=t)
            custom_f1_metric.update(test_preds, test_labels)
            custom_f1_metric_score = custom_f1_metric.compute().item()

            # Check if the current score is the best we've seen
            if custom_f1_metric_score > fmax:
                fmax = custom_f1_metric_score
                best_threshold = t

        return fmax, best_threshold


class Main:
    def evaluate(self, eval_dir: str):
        EvaluatePredictions(eval_dir).evaluate()


if __name__ == "__main__":
    # evaluate_predictions.py evaluate <path/to/file>
    CLI(Main)



import json
import os
from typing import Any, Dict, List, Literal, Union

import torch
from lightning.pytorch.callbacks import BasePredictionWriter


class ChebaiPredictionWriter(BasePredictionWriter):
    """
    A custom prediction writer for saving batch and epoch predictions during model training.

    This class inherits from `BasePredictionWriter` and is designed to save predictions
    in a specified output directory at specified intervals.

    Args:
        output_dir (str): The directory where predictions will be saved.
        write_interval (str): The interval at which predictions will be written.
        target_file (str): The name of the file where epoch predictions will be saved (default: "predictions.json").
    """

    def __init__(
        self,
        output_dir: str,
        write_interval: Literal["batch", "epoch", "batch_and_epoch"],
        target_file: str = "predictions.json",
    ) -> None:
        super().__init__(write_interval)
        self.output_dir = output_dir
        self.target_file = target_file

    def write_on_batch_end(
        self,
        trainer: Any,
        pl_module: Any,
        prediction: Union[torch.Tensor, List[torch.Tensor]],
        batch_indices: List[int],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        """
        Saves batch predictions at the end of each batch.

        Args:
            trainer (Any): The trainer instance.
            pl_module (Any): The LightningModule instance.
            prediction (Union[torch.Tensor, List[torch.Tensor]]): The prediction output from the model.
            batch_indices (List[int]): The indices of the batch.
            batch (Any): The current batch.
            batch_idx (int): The index of the batch.
            dataloader_idx (int): The index of the dataloader.
        """
        outpath = os.path.join(self.output_dir, str(dataloader_idx), f"{batch_idx}.pt")
        os.makedirs(os.path.dirname(outpath), exist_ok=True)
        torch.save(prediction, outpath)

    def write_on_epoch_end(
        self,
        trainer: Any,
        pl_module: Any,
        predictions: List[Dict[str, Any]],
        batch_indices: List[int],
    ) -> None:
        """
        Saves all predictions at the end of each epoch in a JSON file.

        Args:
            trainer (Any): The trainer instance.
            pl_module (Any): The LightningModule instance.
            predictions (List[Dict[str, Any]]): The list of prediction outputs from the model.
            batch_indices (List[int]): The indices of the batches.
        """
        pred_list = []
        for p in predictions:
            idents = p["data"]["idents"]
            labels = p["data"]["labels"]
            if labels is not None:
                labels = labels.tolist()
            else:
                labels = [None for _ in idents]
            output = torch.sigmoid(p["output"]["logits"]).tolist()
            for i, l, o in zip(idents, labels, output):  # noqa: E741
                pred_list.append(dict(ident=i, labels=l, predictions=o))
        with open(os.path.join(self.output_dir, self.target_file), "wt") as fout:
            json.dump(pred_list, fout)



from typing import Dict, List, Tuple, Union

import torch
from torch.nn.utils.rnn import pad_sequence

from chebai.preprocessing.structures import XYData


class Collator:
    """Base class for collating data samples into a batch."""

    def __init__(self, **kwargs):
        pass

    def __call__(self, data: List[Dict]) -> XYData:
        """Collate a list of data samples into a batch.

        Args:
            data (List[Dict]): List of data samples.

        Returns:
            XYData: Batched data.
        """
        raise NotImplementedError


class DefaultCollator(Collator):
    """Default collator that extracts features and labels."""

    def __call__(self, data: List[Dict]) -> XYData:
        """Collate data samples by extracting features and labels.

        Args:
            data (List[Dict]): List of data samples.

        Returns:
            XYData: Batched data.
        """
        x, y = zip(*((d["features"], d["labels"]) for d in data))
        return XYData(x, y)


class RaggedCollator(Collator):
    """
    Collator for handling ragged data samples, designed to support scenarios where some labels may be missing (None).

    This class is specifically designed for preparing batches of "ragged" data, where the samples may have varying sizes,
    such as molecular representations or variable-length protein sequences. Additionally, it supports cases where some
    of the data samples might be partially labeled, which is useful for certain loss functions that allow training
    with incomplete or fuzzy data (e.g., fuzzy loss).

    During batching, the class pads the data samples to a uniform length, applies appropriate masks to differentiate
    between valid and padded elements, and ensures that label misalignment is handled by filtering out unlabelled
    data points. The indices of valid labels are stored in the `non_null_labels` field, which can be used later for
    metrics computation such as F1-score or MSE, especially in cases where some data points lack labels.

    Reference: https://github.com/ChEB-AI/python-chebai/pull/48#issuecomment-2324393829
    """

    def __call__(self, data: List[Union[Dict, Tuple]]) -> XYData:
        """
        Collate ragged data samples (i.e., samples of unequal size, such as molecular sequences) into a batch.

        Handles both fully and partially labeled data, where some samples may have `None` as their label. The indices
        of non-null labels are stored in the `non_null_labels` field, which is used to filter out predictions for
        unlabeled data during evaluation (e.g., F1, MSE). For models supporting partially labeled data, this method
        ensures alignment between features and labels. Missing labels are passed as a loss keyword.

        Args:
            data (List[Union[Dict, Tuple]]): List of ragged data samples. Each sample can be a dictionary or tuple
            with 'features', 'labels', and 'ident'.

        Returns:
            XYData: A batch of padded sequences and labels, including masks for valid positions and indices of
            non-null labels for metric computation.
        """
        model_kwargs: Dict = dict()
        # Indices of non-null labels are stored in key `non_null_labels` of loss_kwargs.
        loss_kwargs: Dict = dict()

        if isinstance(data[0], tuple):
            # For legacy data
            x, y, idents = zip(*data)
            missing_labels = None
        else:
            x, y, idents = zip(
                *((d["features"], d["labels"], d.get("ident")) for d in data)
            )
            missing_labels = [
                d.get("missing_labels", [False for _ in y[0]]) for d in data
            ]

        if any(x is not None for x in y):
            # If any label is not None: (None, None, `1`, None)
            if any(x is None for x in y):
                # If any label is None: (`None`, `None`, 1, `None`)
                non_null_labels = [i for i, r in enumerate(y) if r is not None]
                y = self.process_label_rows(
                    tuple(ye for i, ye in enumerate(y) if i in non_null_labels)
                )
                loss_kwargs["non_null_labels"] = non_null_labels
            else:
                # If all labels are not None: (`0`, `2`, `1`, `3`)
                y = self.process_label_rows(y)

        else:
            # If all labels are None : (`None`, `None`, `None`, `None`)
            y = None
            loss_kwargs["non_null_labels"] = []

        loss_kwargs["missing_labels"] = torch.tensor(missing_labels)
        # Calculate the lengths of each sequence, create a binary mask for valid (non-padded) positions
        lens = torch.tensor(list(map(len, x)))
        model_kwargs["mask"] = torch.arange(max(lens))[None, :] < lens[:, None]
        model_kwargs["lens"] = lens

        return XYData(
            pad_sequence([torch.tensor(a) for a in x], batch_first=True),
            y,
            model_kwargs=model_kwargs,
            loss_kwargs=loss_kwargs,
            idents=idents,
        )

    def process_label_rows(self, labels: Tuple) -> torch.Tensor:
        """
        Process label rows by padding sequences to ensure uniform shape across the batch.

        This method pads the label rows, converting sequences of labels of different lengths into a uniform tensor.
        It ensures that `None` values in the labels are handled by substituting them with a default value(e.g.,`False`).

        Args:
            labels (Tuple): Tuple of label rows.

        Returns:
            torch.Tensor: Padded label sequences.
        """
        return pad_sequence(
            [
                torch.tensor([bool(v) if v is not None else False for v in row])
                for row in labels
            ],
            batch_first=True,
        )



import csv
import gzip
import os
import shutil
from tempfile import NamedTemporaryFile
from typing import Dict, List
from urllib import request

import numpy as np
import torch
from sklearn.model_selection import GroupShuffleSplit, train_test_split

from chebai.preprocessing import reader as dr
from chebai.preprocessing.datasets.base import XYBaseDataModule


class ClinTox(XYBaseDataModule):
    """Data module for ClinTox MoleculeNet dataset."""

    HEADERS = [
        "FDA_APPROVED",
        "CT_TOX",
    ]

    @property
    def _name(self) -> str:
        """Returns the name of the dataset."""
        return "ClinTox"

    @property
    def label_number(self) -> int:
        """Returns the number of labels."""
        return 2

    @property
    def raw_file_names(self) -> List[str]:
        """Returns a list of raw file names."""
        return ["clintox.csv"]

    # @property
    # def processed_file_names(self) -> List[str]:
    #     """Returns a list of processed file names."""
    #     return ["test.pt", "train.pt", "validation.pt"]

    @property
    def processed_file_names_dict(self) -> dict:
        return {
            "test": "test.pt",
            "train": "train.pt",
            "validation": "validation.pt",
        }

    def download(self) -> None:
        """Downloads and extracts the dataset."""
        with NamedTemporaryFile("rb") as gout:
            request.urlretrieve(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/clintox.csv.gz",
                gout.name,
            )
            with gzip.open(gout.name) as gfile:
                with open(os.path.join(self.raw_dir, "clintox.csv"), "wt") as fout:
                    fout.write(gfile.read().decode())

    def setup_processed(self) -> None:
        """Processes and splits the dataset."""
        print("Create splits")
        data = list(
            self._load_data_from_file(os.path.join(self.raw_dir, "clintox.csv"))
        )
        groups = np.array([d["group"] for d in data])
        if not all(g is None for g in groups):
            split_size = int(
                len(set(groups)) * (1 - self.test_split - self.validation_split)
            )
            os.makedirs(self.processed_dir, exist_ok=True)
            splitter = GroupShuffleSplit(train_size=split_size, n_splits=1)

            train_split_index, temp_split_index = next(
                splitter.split(data, groups=groups)
            )

            split_groups = groups[temp_split_index]

            splitter = GroupShuffleSplit(
                train_size=int(
                    len(set(split_groups))
                    * (1 - self.test_split - self.validation_split)
                ),
                n_splits=1,
            )
            test_split_index, validation_split_index = next(
                splitter.split(temp_split_index, groups=split_groups)
            )
            train_split = [data[i] for i in train_split_index]
            test_split = [
                d for d in (data[temp_split_index[i]] for i in test_split_index)
            ]
            validation_split = [
                d for d in (data[temp_split_index[i]] for i in validation_split_index)
            ]
        else:
            train_split, test_split = train_test_split(
                data, test_size=self.test_split, shuffle=True
            )
            train_split, validation_split = train_test_split(
                train_split, test_size=self.validation_split, shuffle=True
            )
        for k, split in [
            ("test", test_split),
            ("train", train_split),
            ("validation", validation_split),
        ]:
            print("transform", k)
            torch.save(
                split,
                os.path.join(self.processed_dir, f"{k}.pt"),
            )

    def setup(self, **kwargs) -> None:
        """Sets up the dataset by downloading and processing if necessary."""
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            self.download()
        if any(
            not os.path.isfile(os.path.join(self.processed_dir, f))
            for f in self.processed_file_names
        ):
            self.setup_processed()

        self._after_setup()

    def _set_processed_data_props(self):
        """
        Load processed data and extract metadata.

        Sets:
            - self._num_of_labels: Number of target labels in the dataset.
            - self._feature_vector_size: Maximum feature vector length across all data points.
        """
        pt_file_path = os.path.join(
            self.processed_dir, self.processed_file_names_dict["train"]
        )
        data_pt = torch.load(pt_file_path, weights_only=False)

        self._num_of_labels = len(data_pt[0]["labels"])
        self._feature_vector_size = max(len(d["features"]) for d in data_pt)

    def _load_dict(self, input_file_path: str) -> List[Dict]:
        """Loads data from a CSV file.

        Args:
            input_file_path (str): Path to the CSV file.

        Returns:
            List[Dict]: List of data dictionaries.
        """
        i = 0
        with open(input_file_path, "r") as input_file:
            reader = csv.DictReader(input_file)
            for row in reader:
                i += 1
                smiles = row["smiles"]
                labels = [
                    bool(int(label)) if label else None
                    for label in (row[k] for k in self.HEADERS)
                ]
                # group = int(row["group"])
                yield dict(
                    features=smiles,
                    labels=labels,
                    ident=i,
                    # group=group
                )
                # yield dict(features=smiles, labels=labels, ident=i)
                # yield self.reader.to_data(dict(features=smiles, labels=labels, ident=i))

    def _perform_data_preparation(self, *args, **kwargs) -> None:
        pass


class BBBP(XYBaseDataModule):
    """Data module for ClinTox MoleculeNet dataset."""

    HEADERS = [
        "p_np",
    ]

    @property
    def _name(self) -> str:
        """Returns the name of the dataset."""
        return "BBBP"

    @property
    def label_number(self) -> int:
        """Returns the number of labels."""
        return 1

    @property
    def raw_file_names(self) -> List[str]:
        """Returns a list of raw file names."""
        return ["bbbp.csv"]

    # @property
    # def processed_file_names(self) -> List[str]:
    #     """Returns a list of processed file names."""
    #     return ["test.pt", "train.pt", "validation.pt"]

    @property
    def processed_file_names_dict(self) -> dict:
        return {
            "test": "test.pt",
            "train": "train.pt",
            "validation": "validation.pt",
        }

    def download(self) -> None:
        """Downloads and extracts the dataset."""
        with open(os.path.join(self.raw_dir, "bbbp.csv"), "ab") as dst:
            with request.urlopen(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/BBBP.csv",
            ) as src:
                shutil.copyfileobj(src, dst)

    def setup_processed(self) -> None:
        """Processes and splits the dataset."""
        print("Create splits")
        data = list(self._load_data_from_file(os.path.join(self.raw_dir, "bbbp.csv")))
        groups = np.array([d["group"] for d in data])
        if not all(g is None for g in groups):
            print("Group shuffled")
            split_size = int(
                len(set(groups)) * (1 - self.test_split - self.validation_split)
            )
            os.makedirs(self.processed_dir, exist_ok=True)
            splitter = GroupShuffleSplit(train_size=split_size, n_splits=1)

            train_split_index, temp_split_index = next(
                splitter.split(data, groups=groups)
            )

            split_groups = groups[temp_split_index]

            splitter = GroupShuffleSplit(
                train_size=int(
                    len(set(split_groups))
                    * (1 - self.test_split - self.validation_split)
                ),
                n_splits=1,
            )
            test_split_index, validation_split_index = next(
                splitter.split(temp_split_index, groups=split_groups)
            )
            train_split = [data[i] for i in train_split_index]
            test_split = [
                d
                for d in (data[temp_split_index[i]] for i in test_split_index)
                # if d["original"]
            ]
            validation_split = [
                d
                for d in (data[temp_split_index[i]] for i in validation_split_index)
                # if d["original"]
            ]
        else:
            train_split, test_split = train_test_split(
                data, test_size=self.test_split, shuffle=True
            )
            train_split, validation_split = train_test_split(
                train_split, test_size=self.validation_split, shuffle=True
            )
        for k, split in [
            ("test", test_split),
            ("train", train_split),
            ("validation", validation_split),
        ]:
            print("transform", k)
            torch.save(
                split,
                os.path.join(self.processed_dir, f"{k}.pt"),
            )

    def setup(self, **kwargs) -> None:
        """Sets up the dataset by downloading and processing if necessary."""
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            self.download()
        if any(
            not os.path.isfile(os.path.join(self.processed_dir, f))
            for f in self.processed_file_names
        ):
            self.setup_processed()

        self._after_setup()

    def _set_processed_data_props(self):
        """
        Load processed data and extract metadata.

        Sets:
            - self._num_of_labels: Number of target labels in the dataset.
            - self._feature_vector_size: Maximum feature vector length across all data points.
        """
        pt_file_path = os.path.join(
            self.processed_dir, self.processed_file_names_dict["train"]
        )
        data_pt = torch.load(pt_file_path, weights_only=False)

        self._num_of_labels = len(data_pt[0]["labels"])
        self._feature_vector_size = max(len(d["features"]) for d in data_pt)

    def _load_dict(self, input_file_path: str) -> List[Dict]:
        """Loads data from a CSV file.

        Args:
            input_file_path (str): Path to the CSV file.

        Returns:
            List[Dict]: List of data dictionaries.
        """
        i = 0
        with open(input_file_path, "r") as input_file:
            reader = csv.DictReader(input_file)
            for row in reader:
                i += 1
                smiles = row["smiles"]
                labels = [int(row["p_np"])]
                # group = int(row["group"])
                yield dict(
                    features=smiles,
                    labels=labels,
                    ident=i,
                    # , group=group
                )
                # yield self.reader.to_data(dict(features=smiles, labels=labels, ident=i))

    def _perform_data_preparation(self, *args, **kwargs) -> None:
        pass


class Sider(XYBaseDataModule):
    """Data module for ClinTox MoleculeNet dataset."""

    HEADERS = [
        "Hepatobiliary disorders",
        "Metabolism and nutrition disorders",
        "Product issues",
        "Eye disorders",
        "Investigations",
        "Musculoskeletal and connective tissue disorders",
        "Gastrointestinal disorders",
        "Social circumstances",
        "Immune system disorders",
        "Reproductive system and breast disorders",
        "Neoplasms benign, malignant and unspecified (incl cysts and polyps)",
        "General disorders and administration site conditions",
        "Endocrine disorders",
        "Surgical and medical procedures",
        "Vascular disorders",
        "Blood and lymphatic system disorders",
        "Skin and subcutaneous tissue disorders",
        "Congenital, familial and genetic disorders",
        "Infections and infestations",
        "Respiratory, thoracic and mediastinal disorders",
        "Psychiatric disorders",
        "Renal and urinary disorders",
        "Pregnancy, puerperium and perinatal conditions",
        "Ear and labyrinth disorders",
        "Cardiac disorders",
        "Nervous system disorders",
        "Injury, poisoning and procedural complications",
    ]

    @property
    def _name(self) -> str:
        """Returns the name of the dataset."""
        return "Sider"

    @property
    def label_number(self) -> int:
        """Returns the number of labels."""
        return 27

    @property
    def raw_file_names(self) -> List[str]:
        """Returns a list of raw file names."""
        return ["sider.csv"]

    # @property
    # def processed_file_names(self) -> List[str]:
    #     """Returns a list of processed file names."""
    #     return ["test.pt", "train.pt", "validation.pt"]

    @property
    def processed_file_names_dict(self) -> dict:
        return {
            "test": "test.pt",
            "train": "train.pt",
            "validation": "validation.pt",
        }

    def download(self) -> None:
        """Downloads and extracts the dataset."""
        with NamedTemporaryFile("rb") as gout:
            request.urlretrieve(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/sider.csv.gz",
                gout.name,
            )
            with gzip.open(gout.name) as gfile:
                with open(os.path.join(self.raw_dir, "sider.csv"), "wt") as fout:
                    fout.write(gfile.read().decode())

    def setup_processed(self) -> None:
        """Processes and splits the dataset."""
        print("Create splits")
        data = list(self._load_data_from_file(os.path.join(self.raw_dir, "sider.csv")))
        groups = np.array([d["group"] for d in data])
        if not all(g is None for g in groups):
            split_size = int(
                len(set(groups)) * (1 - self.test_split - self.validation_split)
            )
            os.makedirs(self.processed_dir, exist_ok=True)
            splitter = GroupShuffleSplit(train_size=split_size, n_splits=1)

            train_split_index, temp_split_index = next(
                splitter.split(data, groups=groups)
            )

            split_groups = groups[temp_split_index]

            splitter = GroupShuffleSplit(
                train_size=int(
                    len(set(split_groups))
                    * (1 - self.test_split - self.validation_split)
                ),
                n_splits=1,
            )
            test_split_index, validation_split_index = next(
                splitter.split(temp_split_index, groups=split_groups)
            )
            train_split = [data[i] for i in train_split_index]
            test_split = [
                d
                for d in (data[temp_split_index[i]] for i in test_split_index)
                # if d["original"]
            ]
            validation_split = [
                d
                for d in (data[temp_split_index[i]] for i in validation_split_index)
                # if d["original"]
            ]
        else:
            train_split, test_split = train_test_split(
                data, test_size=self.test_split, shuffle=True
            )
            train_split, validation_split = train_test_split(
                train_split, test_size=self.validation_split, shuffle=True
            )
        for k, split in [
            ("test", test_split),
            ("train", train_split),
            ("validation", validation_split),
        ]:
            print("transform", k)
            torch.save(
                split,
                os.path.join(self.processed_dir, f"{k}.pt"),
            )

    def setup(self, **kwargs) -> None:
        """Sets up the dataset by downloading and processing if necessary."""
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            self.download()
        if any(
            not os.path.isfile(os.path.join(self.processed_dir, f))
            for f in self.processed_file_names
        ):
            self.setup_processed()

        self._after_setup()

    def _set_processed_data_props(self):
        """
        Load processed data and extract metadata.

        Sets:
            - self._num_of_labels: Number of target labels in the dataset.
            - self._feature_vector_size: Maximum feature vector length across all data points.
        """
        pt_file_path = os.path.join(
            self.processed_dir, self.processed_file_names_dict["train"]
        )
        data_pt = torch.load(pt_file_path, weights_only=False)

        self._num_of_labels = len(data_pt[0]["labels"])
        self._feature_vector_size = max(len(d["features"]) for d in data_pt)

    def _load_dict(self, input_file_path: str) -> List[Dict]:
        """Loads data from a CSV file.

        Args:
            input_file_path (str): Path to the CSV file.

        Returns:
            List[Dict]: List of data dictionaries.
        """
        i = 0
        with open(input_file_path, "r") as input_file:
            reader = csv.DictReader(input_file)
            for row in reader:
                i += 1
                smiles = row["smiles"]
                labels = [
                    bool(int(label)) if label else None
                    for label in (row[k] for k in self.HEADERS)
                ]
                # group = row["group"]
                yield dict(
                    features=smiles,
                    labels=labels,
                    ident=i,
                    # , group=group
                )
                # yield self.reader.to_data(dict(features=smiles, labels=labels, ident=i))

    def _perform_data_preparation(self, *args, **kwargs) -> None:
        pass


class Bace(XYBaseDataModule):
    """Data module for ClinTox MoleculeNet dataset."""

    HEADERS = [
        "class",
    ]

    @property
    def _name(self) -> str:
        """Returns the name of the dataset."""
        return "Bace"

    @property
    def label_number(self) -> int:
        """Returns the number of labels."""
        return 1

    @property
    def raw_file_names(self) -> List[str]:
        """Returns a list of raw file names."""
        return ["bace.csv"]

    # @property
    # def processed_file_names(self) -> List[str]:
    #     """Returns a list of processed file names."""
    #     return ["test.pt", "train.pt", "validation.pt"]

    @property
    def processed_file_names_dict(self) -> dict:
        return {
            "test": "test.pt",
            "train": "train.pt",
            "validation": "validation.pt",
        }

    def download(self) -> None:
        """Downloads and extracts the dataset."""
        with open(os.path.join(self.raw_dir, "bace.csv"), "ab") as dst:
            with request.urlopen(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/bace.csv",
            ) as src:
                shutil.copyfileobj(src, dst)

    def setup_processed(self) -> None:
        """Processes and splits the dataset."""
        print("Create splits")
        data = list(self._load_data_from_file(os.path.join(self.raw_dir, "bace.csv")))
        # groups = np.array([d.get("group") for d in data])

        # if not all(g is None for g in groups):
        #     split_size = int(len(set(groups)) * (1 - self.test_split - self.validation_split))
        #     os.makedirs(self.processed_dir, exist_ok=True)
        #     splitter = GroupShuffleSplit(train_size=split_size, n_splits=1)

        #     train_split_index, temp_split_index = next(
        #         splitter.split(data, groups=groups)
        #     )

        #     split_groups = groups[temp_split_index]

        #     splitter = GroupShuffleSplit(
        #         train_size=int(len(set(split_groups)) * (1 - self.test_split - self.validation_split)), n_splits=1
        #     )
        #     test_split_index, validation_split_index = next(
        #         splitter.split(temp_split_index, groups=split_groups)
        #     )
        #     train_split = [data[i] for i in train_split_index]
        #     test_split = [
        #         d
        #         for d in (data[temp_split_index[i]] for i in test_split_index)
        #     ]
        #     validation_split = [
        #         d
        #         for d in (data[temp_split_index[i]] for i in validation_split_index)
        #     ]
        # else:
        train_split, test_split = train_test_split(
            data, test_size=self.test_split, shuffle=True
        )
        train_split, validation_split = train_test_split(
            train_split, test_size=self.validation_split, shuffle=True
        )
        for k, split in [
            ("test", test_split),
            ("train", train_split),
            ("validation", validation_split),
        ]:
            print("transform", k)
            torch.save(
                split,
                os.path.join(self.processed_dir, f"{k}.pt"),
            )

    def setup(self, **kwargs) -> None:
        """Sets up the dataset by downloading and processing if necessary."""
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            self.download()
        if any(
            not os.path.isfile(os.path.join(self.processed_dir, f))
            for f in self.processed_file_names
        ):
            self.setup_processed()

        self._after_setup()

    def _set_processed_data_props(self):
        """
        Load processed data and extract metadata.

        Sets:
            - self._num_of_labels: Number of target labels in the dataset.
            - self._feature_vector_size: Maximum feature vector length across all data points.
        """
        pt_file_path = os.path.join(
            self.processed_dir, self.processed_file_names_dict["train"]
        )
        data_pt = torch.load(pt_file_path, weights_only=False)

        self._num_of_labels = len(data_pt[0]["labels"])
        self._feature_vector_size = max(len(d["features"]) for d in data_pt)

    def _load_dict(self, input_file_path: str) -> List[Dict]:
        """Loads data from a CSV file.

        Args:
            input_file_path (str): Path to the CSV file.

        Returns:
            List[Dict]: List of data dictionaries.
        """
        i = 0
        with open(input_file_path, "r") as input_file:
            reader = csv.DictReader(input_file)
            for row in reader:
                i += 1
                smiles = row["mol"]
                labels = [int(row["Class"])]
                # group = row["group"]
                yield dict(features=smiles, labels=labels, ident=i)  # , group=group
                # yield self.reader.to_data(dict(features=smiles, labels=labels, ident=i))

    def _perform_data_preparation(self, *args, **kwargs) -> None:
        pass


class HIV(XYBaseDataModule):
    """Data module for ClinTox MoleculeNet dataset."""

    HEADERS = [
        "HIV_active",
    ]

    @property
    def _name(self) -> str:
        """Returns the name of the dataset."""
        return "HIV"

    @property
    def label_number(self) -> int:
        """Returns the number of labels."""
        return 1

    @property
    def raw_file_names(self) -> List[str]:
        """Returns a list of raw file names."""
        return ["hiv.csv"]

    # @property
    # def processed_file_names(self) -> List[str]:
    #     """Returns a list of processed file names."""
    #     return ["test.pt", "train.pt", "validation.pt"]

    @property
    def processed_file_names_dict(self) -> dict:
        return {
            "test": "test.pt",
            "train": "train.pt",
            "validation": "validation.pt",
        }

    def download(self) -> None:
        """Downloads and extracts the dataset."""
        with open(os.path.join(self.raw_dir, "hiv.csv"), "ab") as dst:
            with request.urlopen(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/HIV.csv",
            ) as src:
                shutil.copyfileobj(src, dst)

    def setup_processed(self) -> None:
        """Processes and splits the dataset."""
        print("Create splits")
        data = list(self._load_data_from_file(os.path.join(self.raw_dir, "hiv.csv")))
        groups = np.array([d["group"] for d in data])
        if not all(g is None for g in groups):
            print("Group shuffled")
            split_size = int(
                len(set(groups)) * (1 - self.test_split - self.validation_split)
            )
            os.makedirs(self.processed_dir, exist_ok=True)
            splitter = GroupShuffleSplit(train_size=split_size, n_splits=1)

            train_split_index, temp_split_index = next(
                splitter.split(data, groups=groups)
            )

            split_groups = groups[temp_split_index]

            splitter = GroupShuffleSplit(
                train_size=int(
                    len(set(split_groups))
                    * (1 - self.test_split - self.validation_split)
                ),
                n_splits=1,
            )
            test_split_index, validation_split_index = next(
                splitter.split(temp_split_index, groups=split_groups)
            )
            train_split = [data[i] for i in train_split_index]
            test_split = [
                d for d in (data[temp_split_index[i]] for i in test_split_index)
            ]
            validation_split = [
                d for d in (data[temp_split_index[i]] for i in validation_split_index)
            ]
        else:
            train_split, test_split = train_test_split(
                data, test_size=self.test_split, shuffle=True
            )
            train_split, validation_split = train_test_split(
                train_split, test_size=self.validation_split, shuffle=True
            )
        for k, split in [
            ("test", test_split),
            ("train", train_split),
            ("validation", validation_split),
        ]:
            print("transform", k)
            torch.save(
                split,
                os.path.join(self.processed_dir, f"{k}.pt"),
            )

    def setup(self, **kwargs) -> None:
        """Sets up the dataset by downloading and processing if necessary."""
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            self.download()
        if any(
            not os.path.isfile(os.path.join(self.processed_dir, f))
            for f in self.processed_file_names
        ):
            self.setup_processed()

        self._after_setup()

    def _set_processed_data_props(self):
        """
        Load processed data and extract metadata.

        Sets:
            - self._num_of_labels: Number of target labels in the dataset.
            - self._feature_vector_size: Maximum feature vector length across all data points.
        """
        pt_file_path = os.path.join(
            self.processed_dir, self.processed_file_names_dict["train"]
        )
        data_pt = torch.load(pt_file_path, weights_only=False)

        self._num_of_labels = len(data_pt[0]["labels"])
        self._feature_vector_size = max(len(d["features"]) for d in data_pt)

    def _load_dict(self, input_file_path: str) -> List[Dict]:
        """Loads data from a CSV file.

        Args:
            input_file_path (str): Path to the CSV file.

        Returns:
            List[Dict]: List of data dictionaries.
        """
        i = 0
        with open(input_file_path, "r") as input_file:
            reader = csv.DictReader(input_file)
            for row in reader:
                if len(row) > 1:
                    i += 1
                    smiles = row["smiles"]
                    labels = [int(row["HIV_active"])]
                    # group = int(row["group"])
                    yield dict(
                        features=smiles,
                        labels=labels,
                        ident=i,
                        # , group=group
                    )
                # yield self.reader.to_data(dict(features=smiles, labels=labels, ident=i))

    def _perform_data_preparation(self, *args, **kwargs) -> None:
        pass


class MUV(XYBaseDataModule):
    """Data module for ClinTox MoleculeNet dataset."""

    HEADERS = [
        "MUV-466",
        "MUV-548",
        "MUV-600",
        "MUV-644",
        "MUV-652",
        "MUV-689",
        "MUV-692",
        "MUV-712",
        "MUV-713",
        "MUV-733",
        "MUV-737",
        "MUV-810",
        "MUV-832",
        "MUV-846",
        "MUV-852",
        "MUV-858",
        "MUV-859",
    ]

    @property
    def _name(self) -> str:
        """Returns the name of the dataset."""
        return "MUV"

    @property
    def label_number(self) -> int:
        """Returns the number of labels."""
        return 17

    @property
    def raw_file_names(self) -> List[str]:
        """Returns a list of raw file names."""
        return ["muv.csv"]

    # @property
    # def processed_file_names(self) -> List[str]:
    #     """Returns a list of processed file names."""
    #     return ["test.pt", "train.pt", "validation.pt"]

    @property
    def processed_file_names_dict(self) -> dict:
        return {
            "test": "test.pt",
            "train": "train.pt",
            "validation": "validation.pt",
        }

    def download(self) -> None:
        """Downloads and extracts the dataset."""
        with NamedTemporaryFile("rb") as gout:
            request.urlretrieve(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/muv.csv.gz",
                gout.name,
            )
            with gzip.open(gout.name) as gfile:
                with open(os.path.join(self.raw_dir, "muv.csv"), "wt") as fout:
                    fout.write(gfile.read().decode())

    def setup_processed(self) -> None:
        """Processes and splits the dataset."""
        print("Create splits")
        data = list(self._load_data_from_file(os.path.join(self.raw_dir, "muv.csv")))
        groups = np.array([d["group"] for d in data])
        if not all(g is None for g in groups):
            split_size = int(
                len(set(groups)) * (1 - self.test_split - self.validation_split)
            )
            os.makedirs(self.processed_dir, exist_ok=True)
            splitter = GroupShuffleSplit(train_size=split_size, n_splits=1)

            train_split_index, temp_split_index = next(
                splitter.split(data, groups=groups)
            )

            split_groups = groups[temp_split_index]

            splitter = GroupShuffleSplit(
                train_size=int(
                    len(set(split_groups))
                    * (1 - self.test_split - self.validation_split)
                ),
                n_splits=1,
            )
            test_split_index, validation_split_index = next(
                splitter.split(temp_split_index, groups=split_groups)
            )
            train_split = [data[i] for i in train_split_index]
            test_split = [
                d
                for d in (data[temp_split_index[i]] for i in test_split_index)
                # if d["original"]
            ]
            validation_split = [
                d
                for d in (data[temp_split_index[i]] for i in validation_split_index)
                # if d["original"]
            ]
        else:
            train_split, test_split = train_test_split(
                data, test_size=self.test_split, shuffle=True
            )
            train_split, validation_split = train_test_split(
                train_split, test_size=self.validation_split, shuffle=True
            )
        for k, split in [
            ("test", test_split),
            ("train", train_split),
            ("validation", validation_split),
        ]:
            print("transform", k)
            torch.save(
                split,
                os.path.join(self.processed_dir, f"{k}.pt"),
            )

    def setup(self, **kwargs) -> None:
        """Sets up the dataset by downloading and processing if necessary."""
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            self.download()
        if any(
            not os.path.isfile(os.path.join(self.processed_dir, f))
            for f in self.processed_file_names
        ):
            self.setup_processed()

        self._after_setup()

    def _set_processed_data_props(self):
        """
        Load processed data and extract metadata.

        Sets:
            - self._num_of_labels: Number of target labels in the dataset.
            - self._feature_vector_size: Maximum feature vector length across all data points.
        """
        pt_file_path = os.path.join(
            self.processed_dir, self.processed_file_names_dict["train"]
        )
        data_pt = torch.load(pt_file_path, weights_only=False)

        self._num_of_labels = len(data_pt[0]["labels"])
        self._feature_vector_size = max(len(d["features"]) for d in data_pt)

    def _load_dict(self, input_file_path: str) -> List[Dict]:
        """Loads data from a CSV file.

        Args:
            input_file_path (str): Path to the CSV file.

        Returns:
            List[Dict]: List of data dictionaries.
        """
        i = 0
        with open(input_file_path, "r") as input_file:
            reader = csv.DictReader(input_file)
            for row in reader:
                i += 1
                smiles = row["smiles"]
                labels = [
                    bool(int(label)) if label else None
                    for label in (row[k] for k in self.HEADERS)
                ]
                # group = row["group"]
                yield dict(features=smiles, labels=labels, ident=i)  # , group=group)
                # yield self.reader.to_data(dict(features=smiles, labels=labels, ident=i))

    def _perform_data_preparation(self, *args, **kwargs) -> None:
        pass


class BaceChem(Bace):
    """Chemical data reader for Tox21MolNet dataset."""

    READER = dr.ChemDataReader


class SiderChem(Sider):
    """Chemical data reader for Tox21MolNet dataset."""

    READER = dr.ChemDataReader


class BBBPChem(BBBP):
    """Chemical data reader for Tox21MolNet dataset."""

    READER = dr.ChemDataReader


class ClinToxChem(ClinTox):
    """Chemical data reader for Tox21MolNet dataset."""

    READER = dr.ChemDataReader


class HIVChem(HIV):
    """Chemical data reader for Tox21MolNet dataset."""

    READER = dr.ChemDataReader


class MUVChem(MUV):
    """Chemical data reader for Tox21MolNet dataset."""

    READER = dr.ChemDataReader
